"""
Write pseudo-trajectories to files.

This module contains PtWriter object to write Pseudotrajectories to trajectory/topology files and a PtIOManager that
is a high-level function combining: parsing from input files, creating grids and writing the outputs.
"""
import os

# noinspection PyPep8Naming
import MDAnalysis as mda
from MDAnalysis import Merge
import numpy as np

from molgri.constants import EXTENSION_TRAJECTORY, EXTENSION_TOPOLOGY, EXTENSION_LOGGING
from molgri.logfiles import PtLogger, paths_free_4_all
from molgri.space.fullgrid import FullGrid
from molgri.molecules.parsers import FileParser, ParsedMolecule
from molgri.paths import PATH_INPUT_BASEGRO, PATH_OUTPUT_PT, PATH_OUTPUT_LOGGING
from molgri.molecules.pts import Pseudotrajectory
from molgri.wrappers import time_method


class PtIOManager:

    def __init__(self, name_central_molecule: str, name_rotating_molecule: str, o_grid_name: str, b_grid_name: str,
                 t_grid_name: str, output_name: str = None):
        """
        This class gets only strings as inputs - what a user is expected to provide. The first two strings
        determine where to find input molecular structures, last three how to construct a full grid.
        This object manages Parsers and Writers that provide a smooth input/output to a Pseudotrajectory.

        Args:
            name_central_molecule: name of the molecule that stays fixed (with or without extension,
                                   should be located in the input/ folder)
            name_rotating_molecule: name of the molecule that moves in a pseudotrajectory (with or without extension,
                                   should be located in the input/ folder)
            o_grid_name: name of the grid for rotations around the origin in form 'algorithm_num' (eg. 'ico_50) OR
                         in form 'num' (eg. '50') OR as string 'zero' or 'None' if no rotations needed
            b_grid_name: name of the grid for rotations around the body in form 'algorithm_num' (eg. 'ico_50) OR
                         in form 'num' (eg. '50') OR as string 'zero' or 'None' if no rotations needed
            t_grid_name: translation grid that will be forwarded to TranslationParser, can be a list of numbers,
                         a range or linspace function inside a string
            output_name: select the name under which all associated files will be saved (if None use names of molecules)
        """
        # parsing input files
        central_file_path = f"{PATH_INPUT_BASEGRO}{name_central_molecule}"
        self.central_parser = FileParser(central_file_path)
        rotating_file_path = f"{PATH_INPUT_BASEGRO}{name_rotating_molecule}"
        self.rotating_parser = FileParser(rotating_file_path)
        self.rotating_molecule = self.rotating_parser.as_parsed_molecule()
        self.central_molecule = self.central_parser.as_parsed_molecule()
        # parsing grids
        self.full_grid = FullGrid(b_grid_name=b_grid_name, o_grid_name=o_grid_name, t_grid_name=t_grid_name)
        # initiating writer and pseudotrajectory objects
        self.writer = PtWriter(name_to_save=self.determine_pt_name(),
                               parsed_central_molecule=self.central_molecule)
        self.pt = Pseudotrajectory(self.rotating_molecule, self.full_grid)
        # if the user doesn't select a name, use names of the two molecules as the default
        if output_name is None:
            name_c_molecule = self.central_parser.get_topology_file_name()
            name_r_molecule = self.rotating_parser.get_topology_file_name()
            output_name = f"{name_c_molecule}_{name_r_molecule}"
        self.output_paths = None
        self.output_name = output_name


    def get_decorator_name(self) -> str:
        return f"Pt {self.determine_pt_name()}"

    def get_name(self):
        output_paths = self._get_all_output_paths()
        head, tail = os.path.split(output_paths[0])
        name, ext = os.path.splitext(tail)
        return name

    def determine_pt_name(self) -> str:
        """
        Determine the base name of pseudotrajectory file/directory without any paths or extensions.

        Returns:
            PT name, eg H2O_CL_o_ico_15_b_cube3D_45_t_123456
        """
        name_c_molecule = self.central_parser.get_topology_file_name()
        name_r_molecule = self.rotating_parser.get_topology_file_name()
        name_full_grid = self.full_grid.get_name()
        return f"{name_c_molecule}_{name_r_molecule}_{name_full_grid}"

    def _get_all_output_paths(self, extension_trajectory: str = EXTENSION_TRAJECTORY,
                              extension_structure: str = EXTENSION_TOPOLOGY) -> tuple:
        """
        Return paths to (trajectory_file, structure_file, log_file) with unique number ID.
        """
        if self.output_paths is None:
            # determine the first free file name
            paths = [PATH_OUTPUT_PT, PATH_OUTPUT_PT, PATH_OUTPUT_LOGGING]
            names = [self.output_name]*3
            endings = [extension_trajectory, extension_structure, EXTENSION_LOGGING]
            self.output_paths = paths_free_4_all(list_paths=paths, list_names=names, list_endings=endings)
        return self.output_paths

    def construct_pt(self, extension_trajectory: str = EXTENSION_TRAJECTORY,
                     extension_structure: str = EXTENSION_TOPOLOGY,
                     as_dir: bool = False, print_messages=False):
        """
        The highest-level method to be called in order to generate and save a pseudotrajectory.

        Args:
            extension_trajectory: what extension to provide to the trajectory file
            extension_structure: what extension to provide to the structure (topology) file
            as_dir: if True, don't save trajectory in one file but split it in frames
        """
        path_t, path_s, path_l = self._get_all_output_paths(extension_trajectory=extension_trajectory,
                                                            extension_structure=extension_structure)
        pt_logger = PtLogger(path_l)
        pt_logger.log_set_up(self)
        logger = pt_logger.logger
        logger.info(f"central molecule: {self.central_parser.get_topology_file_name()}")
        logger.info(f"rotating molecule: {self.rotating_parser.get_topology_file_name()}")
        logger.info(f"full grid name: {self.pt.get_full_grid().get_name()}")
        logger.info(f"translation grid [A]: {self.pt.get_full_grid().get_radii()}")
        logger.info(f"quaternions for rot around the body: {self.pt.get_full_grid().get_body_rotations().as_quat()}")
        logger.info(f"positions on a sphere for origin rot: {self.pt.get_full_grid().o_positions}")
        # log set-up before calculating PT in case any errors occur in-between
        if print_messages:
            print(f"Saved the log file to {path_l}")
        # generate a pt
        if as_dir:
            selected_function = self.writer.write_frames_in_directory
        else:
            selected_function = self.writer.write_full_pt
        selected_function(self.pt, path_t, path_s)
        if print_messages:
            print(f"Saved pseudo-trajectory to {path_t} and structure file to {path_s}")


    @time_method
    def construct_pt_and_time(self, **kwargs):
        """
        Same as construct_pt, but time the execution and write out a message about duration.

        Args:
            see construct_pt
        """
        self.construct_pt(**kwargs)


def _create_dir_or_empty_it(directory_name):
    """
    Helper function that determines the name of the directory in which single frames of the trajectory are
    saved. If the directory already exists, its previous contents are deleted.

    Returns:
        path to the directory
    """
    try:
        os.mkdir(directory_name)
    except FileExistsError:
        # delete contents if folder already exist
        filelist = [f for f in os.listdir(directory_name)]
        for f in filelist:
            os.remove(os.path.join(directory_name, f))


class PtWriter:

    def __init__(self, name_to_save: str, parsed_central_molecule: ParsedMolecule):
        """
        This class writes a pseudotrajectory to a file. A PT consists of one molecule that is stationary at
        origin and one that moves with every time step. The fixed molecule is provided when the class is created
        and the mobile molecule as a generator when the method write_full_pt is called. Writing is done with
        MDAnalysis module, so all formats implemented there are supported.

        Args:
            name_to_save: base name of the PT file without paths or extensions
            parsed_central_molecule: a ParsedMolecule object describing the central molecule, will only be translated
                                     so that COM lies at (0, 0, 0) but not manipulated in any other way.
        """
        self.central_molecule = parsed_central_molecule
        self.central_molecule.translate_to_origin()
        self.box = self.central_molecule.get_box()
        self.file_name = name_to_save

    def _merge_and_write(self, writer: mda.Writer, pt: Pseudotrajectory):
        """
        Helper function to merge Atoms from central molecule with atoms of the moving molecule (at current positions)

        Args:
            writer: an already initiated object writing to file (eg a .gro or .xtc file)
            pt: a Pseudotrajectory object with method .get_molecule() that returns current ParsedMolecule
        """
        merged_universe = Merge(self.central_molecule.get_atoms(), pt.get_molecule().get_atoms())
        merged_universe.dimensions = self.box
        writer.write(merged_universe)

    def write_structure(self, pt: Pseudotrajectory, path_structure: str):
        """
        Write the one-frame topology file, eg in .gro format.

        Args:
            pt: a Pseudotrajectory object with method .get_molecule() that returns current ParsedMolecule
            path_structure: where topology should be saved
        """
        if not np.all(self.box == pt.get_molecule().get_box()):
            print(f"Warning! Simulation boxes of both molecules are different. Selecting the box of "
                  f"central molecule with dimensions {self.box}")
        with mda.Writer(path_structure) as structure_writer:
            self._merge_and_write(structure_writer, pt)

    def write_full_pt(self, pt: Pseudotrajectory, path_trajectory: str, path_structure: str):
        """
        Write the trajectory file as well as the structure file (only at the first time-step).

        Args:
            pt: a Pseudotrajectory object with method .generate_pseudotrajectory() that generates ParsedMolecule objects
            path_trajectory: where trajectory should be saved
            path_structure: where topology should be saved
        """
        trajectory_writer = mda.Writer(path_trajectory, multiframe=True)
        last_i = 0
        for i, _ in pt.generate_pseudotrajectory():
            # for the first frame write out topology
            if i == 0:
                self.write_structure(pt, path_structure)
            self._merge_and_write(trajectory_writer, pt)
            last_i = i
        product_of_grids = pt.position_grid.shape[0] * len(pt.rot_grid_body) * pt.position_grid.shape[1]
        assert last_i + 1 == product_of_grids, f"Length of PT not correct, {last_i}=/={product_of_grids}"
        trajectory_writer.close()

    def write_frames_in_directory(self, pt: Pseudotrajectory, path_trajectory: str, path_structure: str):
        """
        As an alternative to saving a full PT in a single trajectory file, you can create a directory with the same
        name and within it single-frame trajectories named with their frame index. Also save the structure file at
        first step.

            pt: a Pseudotrajectory object with method .generate_pseudotrajectory() that generates ParsedMolecule objects
            path_trajectory: where trajectory should be saved
            path_structure: where topology should be saved
        """
        directory_name, extension_trajectory = os.path.splitext(path_trajectory)
        _create_dir_or_empty_it(directory_name)
        for i, _ in pt.generate_pseudotrajectory():
            if i == 0:
                self.write_structure(pt, path_structure)
            f = f"{directory_name}/{i}.{extension_trajectory}"
            with mda.Writer(f) as structure_writer:
                self._merge_and_write(structure_writer, pt)



"""
NetworkX graphs representing Platonic solids in 3- or 4D.

Specifically, we implement the 3D polytopes
icosahedron and cube and the 4D polytope hypercube. As a start, vertices of the polytope are added as
nodes and edges as connections between them. For the cube, a point is also added in the middle of each face diagonal
and the edges from these face points to the four vertices are added. For the hypercube, a point is added at the center
of each of the 8 sub-cells and in the middle of all square faces, edges to these points are also added.

It's important that polytopes can also be subdivided in smaller units so that grids with larger numbers of points
can be created. For this, use divide_edges() command. 3D polytopes (polyhedra) can also be plotted - for 4D polyhedra,
each 3D cell can be plotted.

Points and their projections should always be accessed with getters and include only non-repeating rotation
quaternions for cube4D.

Objects:
    - Polytope (abstract)
    - Polyhedron (abstract, only 3D, implements plotting)
    - PolyhedronFromG (helper for testing/plotting, initiated with an existing graph)
    - Cube4DPolytope
    - IcosahedronPolytope
    - Cube3DPolytope
"""

from abc import ABC, abstractmethod
from itertools import product, combinations
from typing import Hashable, Iterable

import networkx as nx
import numpy as np
from numpy.typing import NDArray
from scipy.constants import pi, golden

from molgri.assertions import is_array_with_d_dim_r_rows_c_columns, all_row_norms_equal_k, form_square_array, form_cube
from molgri.space.utils import normalise_vectors, dist_on_sphere, unique_quaternion_set


class Polytope(ABC):

    """
    A polytope is a d-dim object consisting of a set of nodes (vertices) and connections between them (edges) saved
    in self.G (graph).

    The basic polytope will be created when the object is initiated. All further divisions should be performed
    with the function self.divide_edges()
    """

    def __init__(self, d: int = 3, is_quat: bool = False):
        """
        Args:
            d: number of dimensions
            is_quat: does the object represent quaternions (meaning that q and -q points are treated as equal)
        """

        self.G = nx.Graph()
        self.current_level = 0
        self.side_len = 0
        self.d = d
        self.is_quat = is_quat

        # saved objects in order to not re-construct them unnecessarily (unless the level of division has changed)
        self.current_nodes = (None, self.current_level)
        self.current_projections = (None, self.current_level)

        # function that depends on the object being created
        self._create_level0()

    def __str__(self):
        return "Polytope"

    @abstractmethod
    def _create_level0(self):
        """This is implemented by each subclass since they have different edges, vertices and faces"""

    def get_node_coordinates(self) -> NDArray:
        """
        Get an array in which each row represents the self.d - dimensional coordinates of the node. These are the
        points on the faces of the polytope, their norm may not be one.

        If the object is_quat, approximately half of the nodes will be returned, since for quaternions q=-q.

        Returns:
            a numpy array in which each row is one of the nodes in order of their addition to the graph (not sorted)
        """
        # check if saved available and the level of division has not changed since saving
        if self.current_nodes[0] is not None and self.current_level == self.current_nodes[1]:
            # noinspection PyTypeChecker
            return self.current_nodes[0]
        nodes = self.G.nodes
        nodes = np.array(nodes)
        if self.is_quat:
            nodes = unique_quaternion_set(nodes)
        # save nodes for further use:
        self.current_nodes = (nodes, self.current_level)
        return nodes

    def get_projection_coordinates(self) -> NDArray:
        """
        Get an array in which each row represents the self.d - dimensional coordinates of the node projected on a
        self.d-dimensional unit sphere. These points must have norm 1.

        If the object is_quat, approximately half of the nodes will be returned, since for quaternions q=-q.

        Returns:
            a numpy array in which each row is one of the projections in order of their addition to the graph
            (not sorted)
        """
        # check if saved available and the level of division has not changed since saving
        if self.current_projections[0] is not None and self.current_level == self.current_projections[1]:
            # noinspection PyTypeChecker
            return self.current_projections[0]
        projections_dict = nx.get_node_attributes(self.G, "projection")
        # important to rely on the getter because it takes care of is_quat
        nodes = self.get_node_coordinates()
        projections = np.zeros((len(nodes), self.d))
        for i, n in enumerate(nodes):
            projections[i] = projections_dict[tuple(n)]

        # assertions
        is_array_with_d_dim_r_rows_c_columns(projections, d=2, r=len(nodes), c=self.d)
        all_row_norms_equal_k(projections, 1)

        # save projections for further use:
        self.current_projections = (projections, self.current_level)
        return projections

    def get_N_ordered_points(self, N: int = None, projections: bool = True) -> NDArray:
        """
        Get nodes or their projections - this getter implements sorting and truncation at N.

        Args:
            N: wished number of points to return (if None, return all)
            projections: if True, use projections on unit sphere, if False, points on the polytope

        Returns:
            an array of size (N, self.d) featuring N nodes/projections sorted for max distance between them

        Raises:
            ValueError if N larger than the number of available unique nodes
        """

        def remove_and_reconnect(g: nx.Graph, node: tuple):
            """Remove a node and reconnect edges, adding the properties of previous edges together."""
            sources = list(g.neighbors(node))
            targets = list(g.neighbors(node))

            new_edges = list(product(sources, targets))
            distances = [g.edges[(node, s)]["p_dist"]+g.edges[(node, t)]["p_dist"] for s, t in new_edges]
            # remove self-loops
            new_edges = [(edge[0], edge[1], {"p_dist": distances[j]}) for j, edge in enumerate(new_edges) if
                         edge[0] != edge[1]]
            g.add_edges_from(new_edges)
            g.remove_node(node)

        # important to use the getter
        current_points = [tuple(point) for point in self.get_node_coordinates()]
        N_available = len(current_points)
        # can't order more points than there are
        if N is None:
            N = N_available
        if N > N_available:
            raise ValueError(f"Cannot order more points than there are! N={N} > {N_available}")
        result = np.zeros((N, self.d))
        # first point does not matter, just select the first point
        if projections:
            result[0] = self.G.nodes[tuple(current_points[0])]["projection"]
        else:
            result[0] = current_points[0]

        subgraph = self.G.subgraph(current_points).copy()
        remove_and_reconnect(subgraph, current_points[0])
        current_points.pop(0)
        # for the rest of the points, determine centrality of the subgraph with already used points removed
        for i in range(1, N):
            # TODO: use sub-graphs removed_points and remaining_points
            max_iter = np.max([100, 10*N_available])  # bigger graphs may need more iterations than default
            dict_centrality = nx.eigenvector_centrality(subgraph, weight="p_dist", max_iter=max_iter, tol=1.0e-3)
            # key with largest value is the point in the center of the remaining graph
            most_distant_point = min(dict_centrality, key=dict_centrality.get)
            if projections:
                result[i] = self.G.nodes[tuple(most_distant_point)]["projection"]
            else:
                result[i] = most_distant_point
            current_points.remove(most_distant_point)
            remove_and_reconnect(subgraph, most_distant_point)
        return result

    def divide_edges(self):
        """
        Subdivide once by putting a new point at mid-point of each existing edge and replacing this sub-edge with
        two edges from the two old to the new point.

        In sub-modules, additional edges may be added before performing divisions.
        """
        self._add_mid_edge_nodes()
        self.current_level += 1
        self.side_len = self.side_len / 2

    def _add_mid_edge_nodes(self):
        """
        For each edge in the system, add a new point in the middle of the edge and replace the previous long edge with
        two shorter ones.
        """
        nodes_to_add = list(self.G.edges())
        self._add_average_point_and_edges(nodes_to_add)
        # remove the old connection: old_point1 -------- old_point2
        for old1, old2 in nodes_to_add:
            self.G.remove_edge(old1, old2)

    def _add_square_diagonal_nodes(self):
        """
        Detect nodes that form a square. If you find them, add the middle point (average of the four) to nodes along
        with edges from the middle to the points that form the square.
        """
        # here save the new node along with the 4 "father nodes" so you can later build all connections
        square_nodes = detect_all_squares(self.G)
        # add new nodes and edges
        self._add_average_point_and_edges(square_nodes)

    def _add_cube_diagonal_nodes(self):
        """
        Detect nodes that form a cube. If you find them, add the middle point (average of the eight) to nodes along
        with edges from the middle to the points that form the square.
        """
        # here save the new node along with the 4 "father nodes" so you can later build all connections
        cube_nodes = detect_all_cubes(self.G)
        # add new nodes and edges
        self._add_average_point_and_edges(cube_nodes)

    def _add_average_point_and_edges(self, list_old_points: list):
        """
        A helper function to _add_square_diagonal_nodes, _add_cube_diagonal_nodes and _add_mid_edge_nodes.
        Given a list in which every item is a set of nodes, add a new point that is the average of them and also add
        new edges between the newly created points and all of the old ones.

        It adds new edges and does not delete any existing ones.

        Args:
            list_old_points: each item is a list of already existing nodes - their average should be added as
            a new point
        """
        for new_neighbours in list_old_points:
            # new node is the midpoint of the old ones
            new_node_arr = np.average(np.array(new_neighbours), axis=0)
            new_node = tuple(new_node_arr)
            # just in case repetition happens
            if new_node not in self.G.nodes:
                self.G.add_node(new_node, level=self.current_level, face=self._find_face(new_neighbours),
                                projection=normalise_vectors(np.array(new_node)))
                # since an average, only need to calculate distances once
                distance = np.linalg.norm(new_node_arr - np.array(new_neighbours[0]))
                for n in new_neighbours:
                    # IMPORTANT: even if distances on a polyhedron are the same, distances on the sphere no longer
                    # are since the distance of points to the origin depends on where on the face of the polyhedron
                    # the newly connected points appear
                    length = dist_on_sphere(self.G.nodes[new_node]["projection"],
                                            self.G.nodes[n]["projection"])[0]
                    self.G.add_edge(new_node, n, p_dist=distance, length=length)

    def _add_edges_of_len(self, edge_len: float, wished_levels: list[int] = None, only_seconds: bool = True,
                          only_face: bool = True):
        """
        Sometimes, additional edges among already existing points need to be added in order to achieve equal
        sub-division of all surfaces.

        In order to shorten the time to search for appropriate points, it is advisable to make use of filters:
         - only_seconds will only search for connections between points that are second neighbours of each other
         - wished_levels defines at which division level the points between which the edge is created should be
         - only_face: only connections between points on the same face will be created

        Args:
            edge_len: length of edge on the polyhedron surface that is condition for adding edges
            wished_levels: at what level of division should the points that we want to connect be?
            only_seconds: if True, search only among the points that are second neighbours
            only_face: if True, search only among the points that lie on the same face
        """
        if wished_levels is None:
            wished_levels = [self.current_level, self.current_level]
        else:
            wished_levels.sort(reverse=True)
        assert len(wished_levels) == 2
        selected_level = [x for x, y in self.G.nodes(data=True) if y['level'] == wished_levels[0]]
        for new_node in selected_level:
            # searching only second neighbours of the node if this option has been selected
            if only_seconds:
                sec_neighbours = list(second_neighbours(self.G, new_node))
                sec_neighbours = [x for x in sec_neighbours if self.G.nodes[x]["level"] == wished_levels[1]]
            else:
                sec_neighbours = [x for x in self.G.nodes if self.G.nodes[x]["level"] == wished_levels[1]]
            for other_node in sec_neighbours:
                node_dist = np.linalg.norm(np.array(new_node)-np.array(other_node))
                # check face criterion
                if not only_face or self._find_face([new_node, other_node]):
                    # check distance criterion
                    if np.isclose(node_dist, edge_len):
                        length = dist_on_sphere(self.G.nodes[new_node]["projection"],
                                                self.G.nodes[other_node]["projection"])[0]
                        # just to make sure that you don't add same edge in 2 different directions
                        if new_node < other_node:
                            self.G.add_edge(new_node, other_node, p_dist=edge_len, length=length)
                        else:
                            self.G.add_edge(other_node, new_node, p_dist=edge_len, length=length)

    def _find_face(self, node_list: list) -> set:
        """
        Find the face that is common between all nodes in the list. (May be an empty set.)

        Args:
            node_list: a list in which each element is a tuple of coordinates that has been added to self.G as a node

        Returns:
            the set of faces (may be empty) that all nodes in this list share
        """
        face = set(self.G.nodes[node_list[0]]["face"])
        for neig in node_list[1:]:
            faces_neighbour_vector = self.G.nodes[neig]["face"]
            face = face.intersection(set(faces_neighbour_vector))
        return face


class PolyhedronFromG(Polytope):

    """
    A mock polyhedron created from an existing graph. No guarantee that it actually represents a polyhedron. Useful for
    side tasks like visualisations, searches ...
    """

    def __init__(self, G: nx.Graph):
        super().__init__()
        self.G = G

    def _create_level0(self):
        pass


class Cube4DPolytope(Polytope):
    """
    CubeGrid is a graph object, its central feature is self.G (networkx graph). In the beginning, each node is
    a vertex of a 3D cube, its center or the center of its face. It is possible to subdivide the sides, in that case
    the volumes of sub-cubes are fully sub-divided.

    Special: the attribute "face" that in 3D polyhedra actually means face here refers to the cell (3D object) to which
    this specific node belongs.
    """

    def __init__(self):
        super().__init__(d=4, is_quat=True)

    def __str__(self):
        return f"Cube4D up to level {self.current_level}"

    def _add_average_point(self, list_old_points: list):
        """
        A helper function to _add_square_diagonal_nodes, _add_cube_diagonal_nodes and _add_mid_edge_nodes.
        Given a list in which every item is a set of nodes, add a new point that is the average of them and also add
        new edges between the newly created points and all of the old ones.

        It adds new edges and does not delete any existing ones.

        Args:
            list_old_points: each item is a list of already existing nodes - their average should be added as
            a new point
        """
        # TODO: refactor so that this is part of Polyhedron and adding edges is an extra step
        for new_neighbours in list_old_points:
            # new node is the midpoint of the old ones
            new_node_arr = np.average(np.array(new_neighbours), axis=0)
            new_node = tuple(new_node_arr)
            # just in case repetition happens
            if new_node not in self.G.nodes:
                self.G.add_node(new_node, level=self.current_level, face=self._find_face(new_neighbours),
                                projection=normalise_vectors(np.array(new_node)))

    def _add_point_at_len(self, edge_len: float, wished_levels: list = None, only_seconds: bool = True,
                          only_face: bool = True):
        """
        Sometimes, additional edges among already existing points need to be added in order to achieve equal
        sub-division of all surfaces.

       See also Polyhedron's function add_edges_of_len
        """
        # TODO: also join with Polyhedron
        if wished_levels is None:
            wished_levels = [self.current_level, self.current_level]
        else:
            wished_levels.sort(reverse=True)
        assert len(wished_levels) == 2
        selected_level = [x for x, y in self.G.nodes(data=True) if y['level'] == wished_levels[0]]
        for new_node in selected_level:
            # searching only second neighbours of the node if this option has been selected
            if only_seconds:
                sec_neighbours = list(second_neighbours(self.G, new_node))
                sec_neighbours = [x for x in sec_neighbours if self.G.nodes[x]["level"] == wished_levels[1]]
            else:
                sec_neighbours = [x for x in self.G.nodes if self.G.nodes[x]["level"] == wished_levels[1]]
            for other_node in sec_neighbours:
                node_dist = np.linalg.norm(np.array(new_node)-np.array(other_node))
                # check face criterion
                if not only_face or self._find_face([new_node, other_node]):
                    # check distance criterion
                    if np.isclose(node_dist, edge_len):
                        self._add_average_point([[new_node, other_node]])

    def _create_level0(self):
        self.side_len = 2 * np.sqrt(1/self.d)
        # create vertices
        vertices = list(product((-self.side_len/2, self.side_len/2), repeat=4))
        assert len(vertices) == 16
        assert np.all([np.isclose(x, 0.5) or np.isclose(x, -0.5) for row in vertices for x in row])
        assert len(set(vertices)) == 16
        vertices = np.array(vertices)
        # add vertices and edges to the graph
        for i, vert in enumerate(vertices):
            self.G.add_node(tuple(vert), level=self.current_level, face=set(),
                            projection=normalise_vectors(np.array(vert)))
        self._add_edges_of_len(self.side_len, wished_levels=[self.current_level, self.current_level],
                               only_seconds=False, only_face=False)
        # detect cells
        cells = detect_all_cubes(self.G)
        for i, cell in enumerate(cells):
            # find the indices of the vertices that construct this square in the list of vertices
            for vertex in cell:
                self.G.nodes[tuple(vertex)]["face"].add(i)
        self.side_len = self.side_len / 2
        self.current_level += 1

    def divide_edges(self):
        """Before or after dividing edges, make sure all relevant connections are present. Then perform a division of
        all connections (point in the middle, connection split in two), increase the level and halve the side_len."""
        self._add_point_at_len(2*self.side_len*np.sqrt(2), wished_levels=[self.current_level-1, self.current_level-1],
                               only_seconds=False)
        self._add_point_at_len(2 * self.side_len * np.sqrt(3),
                               wished_levels=[self.current_level - 1, self.current_level - 1],
                               only_seconds=False)
        super().divide_edges()

        self._add_edges_of_len(2*self.side_len, wished_levels=[self.current_level-1, self.current_level-1],
                               only_seconds=False)


class IcosahedronPolytope(Polytope):
    """
    IcosahedronPolytope is a graph object, its central feature is self.G (networkx graph). In the beginning, each node
    is a vertex of a 3D icosahedron. It is possible to subdivide the sides, in that case a new point always appears in
    the middle of each triangle side.
    """

    def __init__(self):
        super().__init__(d=3, is_quat=False)

    def __str__(self):
        return f"Icosahedron up to level {self.current_level}"

    def _create_level0(self):
        # DO NOT change order of points - faces will be wrong!
        # each face contains numbers - indices of vertices that border on this face
        faces = [[0, 11, 5], [0, 5, 1], [0, 1, 7], [0, 7, 10], [0, 10, 11], [1, 5, 9], [5, 11, 4], [11, 10, 2],
                 [10, 7, 6], [7, 1, 8], [3, 9, 4], [3, 4, 2], [3, 2, 6], [3, 6, 8], [3, 8, 9], [4, 9, 5], [2, 4, 11],
                 [6, 2, 10], [8, 6, 7], [9, 8, 1]]
        self.side_len = 1 / np.sin(2 * pi / 5)
        # create vertices
        vertices = [(-1, golden, 0), (1, golden, 0), (-1, -golden, 0), (1, -golden, 0),
                    (0, -1, golden), (0, 1, golden), (0, -1, -golden), (0, 1, -golden),
                    (golden, 0, -1), (golden, 0, 1), (-golden, 0, -1), (-golden, 0, 1)]
        vertices = np.array(vertices) * self.side_len / 2
        # add vertices and edges to the graph
        for i, vert in enumerate(vertices):
            set_of_faces = set(faces_i for faces_i, face in enumerate(faces) if i in face)
            self.G.add_node(tuple(vert), level=self.current_level, face=set_of_faces,
                            projection=normalise_vectors(np.array(vert)))
        # create edges
        self._add_edges_of_len(self.side_len, wished_levels=[self.current_level, self.current_level],
                               only_seconds=False, only_face=False)
        # just to check ...
        assert self.G.number_of_nodes() == 12
        assert self.G.number_of_edges() == 30
        for node in self.G.nodes(data=True):
            assert len(node[1]["face"]) == 5 and node[1]["level"] == 0
        self.side_len = self.side_len / 2
        self.current_level += 1

    def divide_edges(self):
        """
        Subdivide once. If previous faces are triangles, adds one point at mid-point of each edge. If they are
        squares, adds one point at mid-point of each edge + 1 in the middle of the face. New points will have a higher
        level attribute.
        """
        # also add edges between second neighbours that form triangles
        self._add_edges_of_len(self.side_len*2, wished_levels=[self.current_level-1, self.current_level-1],
                               only_seconds=True)
        # this positions all new points we need - simply in the middle of previous edges
        super().divide_edges()


class Cube3DPolytope(Polytope):
    """
    In the beginning, each node is a vertex of a 3D cube + 6 vertices at mid-faces. It is possible to subdivide
    the sides, in that case a new point always appears in the middle of a square and half of previous sides.
    """

    def __init__(self):
        super().__init__(d=3, is_quat=False)

    def __str__(self):
        return f"Cube3D up to level {self.current_level}"

    def _create_level0(self):
        # DO NOT change order of points - faces will be wrong!
        # each face contains numbers - indices of vertices that border on this face
        faces = [[0, 1, 2, 4], [0, 2, 3, 6], [0, 1, 3, 5], [3, 5, 6, 7], [1, 4, 5, 7], [2, 4, 6, 7]]
        self.side_len = 2 * np.sqrt(1/3)
        # create vertices
        vertices = [(-self.side_len/2, -self.side_len/2, -self.side_len/2),
                    (-self.side_len/2, -self.side_len/2, self.side_len/2),
                    (-self.side_len/2, self.side_len/2, -self.side_len/2),
                    (self.side_len/2, -self.side_len/2, -self.side_len/2),
                    (-self.side_len/2, self.side_len/2, self.side_len/2),
                    (self.side_len/2, -self.side_len/2, self.side_len/2),
                    (self.side_len/2, self.side_len/2, -self.side_len/2),
                    (self.side_len/2, self.side_len/2, self.side_len/2)]
        vertices = np.array(vertices)

        # add vertices and edges to the graph
        for i, vert in enumerate(vertices):
            set_of_faces = tuple(faces_i for faces_i, face in enumerate(faces) if i in face)
            self.G.add_node(tuple(vert), level=self.current_level, face=set_of_faces,
                            projection=normalise_vectors(np.array(vert)))
        # create edges
        self._add_edges_of_len(self.side_len, wished_levels=[self.current_level, self.current_level],
                               only_seconds=False, only_face=False)
        # create diagonal nodes
        self._add_square_diagonal_nodes()
        self.side_len = self.side_len / 2
        self.current_level += 1

    def divide_edges(self):
        """
        Subdivide each existing edge. Before the division make sure there are diagonal and straight connections so that
        new points appear at mid-edges and mid-faces.
        """
        # also add the diagonals in the other direction
        self._add_edges_of_len(self.side_len * np.sqrt(2),
                               wished_levels=[self.current_level-1, self.current_level-1], only_seconds=True)
        # add the straight lines connecting old points to new
        if self.current_level >= 2:
            self._add_edges_of_len(self.side_len*2, wished_levels=[self.current_level-1, self.current_level-2],
                                   only_seconds=True)
        super().divide_edges()

#######################################################################################################################
#                                          GRAPH HELPER FUNCTIONS
#######################################################################################################################


def second_neighbours(graph: nx.Graph, node: Hashable) -> Iterable:
    """
    Yield second neighbors of node in graph. Ignore second neighbours that are also first neighbours.
    Second neighbors may repeat!

    Example:

        5------6
        |      |
        2 ---- 1 ---- 3 ---- 7
               |      |
               |__8___|

    First neighbours of 1: 2, 6, 3, 8
    Second neighbours of 1: 5, 7
    """
    direct_neighbours = list(graph.neighbors(node))
    # don't repeat the same second neighbour twice
    seen_seconds = []
    for neighbor_list in [graph.neighbors(n) for n in direct_neighbours]:
        for n in neighbor_list:
            if n != node and n not in direct_neighbours and n not in seen_seconds:
                # if no edge there, put it there just to record the distance
                seen_seconds.append(n)
                yield n


def third_neighbours(graph: nx.Graph, node: Hashable) -> Iterable:
    """
    Yield third neighbors of node in graph. Analogous to second neighbours, one more degree of separation

    Example:

        5------6
        |      |
        2 ---- 1 ---- 3 ---- 7 ---- 9
               |      |
               11__8__10

    First neighbours of 1: 2, 6, 3, 11
    Second neighbours of 1: 5, 8, 10, 7
    Third neighbours of 1: 9
    """
    direct_neighbours = list(graph.neighbors(node))
    sec_neighbours = list(second_neighbours(graph, node))
    # don't repeat seen third neighbours either
    third_seen = []
    for neighbor_list in [graph.neighbors(n) for n in sec_neighbours]:
        for n in neighbor_list:
            # conditions: n must not be: the node itself, its first or second neighbour or already seen third neighbour
            if n != node and n not in direct_neighbours and n not in sec_neighbours and n not in third_seen:
                third_seen.append(n)
                yield n


def detect_all_squares(graph: nx.Graph) -> list:
    """
    Each node of the graph is represented by 3- or 4-dimensional coordinates, meaning that distances and even angles
    to other nodes  can be calculated. Squares can be formed by a combination of: a node, 2 neighbours and 1 second
    neighbour. Detect all such occurrences and return a list in which each row is a new set of 4 points forming a
    square.

    Args:
        graph: Graph in which each node is represented by a tuple of numbers representing coordinates

    Returns:
        a list, each item a set of 4 vertices that are guaranteed to form a square
    """
    square_nodes = []
    for node in graph.nodes:
        neighbours = list(graph.neighbors(node))
        sec_neighbours = list(second_neighbours(graph, node))
        # prepare all combinations
        # a possibility for a square: node, two neighbours, one second neighbour
        for sn in sec_neighbours:
            n_choices = list(combinations(neighbours, 2))
            for two_n in n_choices:
                n1, n2 = two_n
                # points sorted by size so that no duplicates occur
                points = sorted([node, sn, n1, n2])
                if form_square_array(np.array(points)) and points not in square_nodes:
                    square_nodes.append([*points])
    return square_nodes


def detect_all_cubes(graph: nx.Graph) -> list:
    """
    See detect_all_squares, but now a cube consists of: a node, 3 direct neighbours, 3 second neighbours, 1 third
    neighbour.

    Args:
        graph: Graph in which each node is represented by a tuple of numbers representing coordinates

    Returns:
        a list, each item a set of 8 vertices that are guaranteed to form a cube
    """
    cube_nodes = []
    for node in graph.nodes:
        neighbours = list(graph.neighbors(node))
        sec_neighbours = list(second_neighbours(graph, node))
        th_neighbours = list(third_neighbours(graph, node))
        # prepare all combinations: a node, 3 direct neighbours, 3 second neighbours, 1 third neighbour
        first_choices = list(combinations(neighbours, 3))
        second_choices = list(combinations(sec_neighbours, 3))
        for tn in th_neighbours:
            for fn in first_choices:
                for sn in second_choices:
                    # points sorted by size so that no duplicates occur
                    points = sorted([node, *fn, *sn, tn])
                    array_points = np.array(points)
                    if form_cube(array_points) and points not in cube_nodes:
                        cube_nodes.append([*points])
    return cube_nodes



"""
Plotting 2D array as heatmap.
"""

import numpy as np
from numpy.typing import NDArray
import seaborn as sns
from matplotlib.pyplot import Figure, Axes
import matplotlib.colors as colors

from molgri.plotting.abstract import RepresentationCollection


class ArrayPlot(RepresentationCollection):

    """
    A tool for plotting arrays, eg by highlighting high and low values
    """

    def __init__(self, my_array: NDArray, *args, data_name="array", **kwargs):
        self.array = my_array
        super().__init__(data_name, *args, **kwargs)

    def make_heatmap_plot(self, ax: Axes = None, fig: Figure = None, save: bool = True):
        """
        This method draws the array and colors the fields according to their values (red = very large,
        blue = very small). Zero values are always white, negative ones always blue, positive ones always red.
        """
        self._create_fig_ax(fig=fig, ax=ax)
        if np.all(self.array < 0):
            cmap = "Blues"
            norm = None
        elif np.all(self.array > 0):
            cmap = "Reds"
            norm = None
        else:
            cmap = "bwr"
            norm = colors.TwoSlopeNorm(vcenter=0)
        sns.heatmap(self.array, cmap=cmap, ax=self.ax, xticklabels=False, yticklabels=False, norm=norm)

        self._equalize_axes()
        if save:
            self._save_plot_type("heatmap")

if __name__ == "__main__":
    import pandas as pd
    from molgri.molecules.transitions import SimulationHistogram, SQRA, MSM
    from molgri.molecules.parsers import ParsedEnergy, FileParser
    from molgri.space.fullgrid import FullGrid

    path_energy = "/home/mdglasius/Modelling/trypsin_test/output/measurements/temp_minimized.csv"
    df = pd.read_csv(path_energy)
    energies = df['potential'].to_numpy()[:, np.newaxis]
    pe = ParsedEnergy(energies=energies, labels=["Potential"], unit="(kJ/mole)")
    pt_parser = FileParser(
        path_topology="/home/mdglasius/Modelling/trypsin_test/output/pt_files/final_trypsin_NH4_o_ico_512_b_zero_1_t_2153868773.gro",
        path_trajectory="/home/mdglasius/Modelling/trypsin_test/output/pt_files/minimizedPT.pdb")
    parsed_trajectory = pt_parser.get_parsed_trajectory(default_atom_selection="not protein")
    parsed_trajectory.energies = pe
    print(len(parsed_trajectory.get_all_COM()))

    # the exact grid used
    fg = FullGrid(t_grid_name="linspace(0.8, 2.5, 4)", o_grid_name="ico_10", b_grid_name="zero")

    # plotting
    sh = SimulationHistogram(parsed_trajectory, fg)
    my_sqra = SQRA(sh, use_saved=False)

    rate_matrix = my_sqra.get_transitions_matrix()

    # visualise rate matrix
    ArrayPlot(rate_matrix[0], data_name="rate_matrix").make_heatmap_plot()

    # TRANSITION MATRIX
    pt_parser = FileParser(
        path_topology="/home/mdglasius/Modelling/trypsin_normal/inputs/trypsin_probe.pdb",
        path_trajectory="/home/mdglasius/Modelling/trypsin_normal/nobackup/outputs/aligned_traj.dcd")
    parsed_trajectory = pt_parser.get_parsed_trajectory(default_atom_selection="segid B")
    print(len(parsed_trajectory.get_all_COM()))

    # preparing the grid
    fg = FullGrid(t_grid_name="linspace(3, 13, 4)", o_grid_name="ico_40", b_grid_name="zero")

    sh = SimulationHistogram(parsed_trajectory, fg)
    my_msm = MSM(sh, use_saved=False)
    transition_matrix = my_msm.get_transitions_matrix()

    # visualise transition matrix
    ArrayPlot(transition_matrix[0], data_name="transition_matrix").make_heatmap_plot()



"""
Full discretisation of space in spherical layers.

The module fullgrid combines a linear translation grid with two spherical grids: discretisation of approach vectors
(3D sphere, orientation grid) and of internal rotations of the second body (4D half-sphere of quaternions, body grid).
The three grids are commonly referred to as t_grid, o_grid and b_grid.

Position grid is a product of t_grid and o_grid and represents a set of spherical points in 3D space that are repeated
at different radii. Based on a position grid it is possible to create a Voronoi discretisation of 3D space using
identical (up to radius) layers of Voronoi surfaces between layers of grid points and connecting them with ray points
from the origin to are vertices of Voronoi cells. Methods to calculate volumes, areas and distances between centers of
such discretisation sof position space are provided.

Objects:
 - FullGrid combines 3D sphere grid, 4D sphere grid and a translation grid
 - FullVoronoiGrid extends FullGrid with methods to evaluate distances/areas/volumes of cells
 - Point represents a single point in FullVoronoiGrid and implements helper functions like identifying vertices of cell
 - ConvergenceFullGridO provides plotting data by creating a range of FullGrids with different N of o_grid points
"""
from __future__ import annotations
from copy import copy
from typing import Callable, Tuple, Optional


import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation
from scipy.spatial import SphericalVoronoi
from scipy.spatial.distance import cdist
from scipy.constants import pi
from scipy.sparse import coo_array, csc_array
import pandas as pd

from molgri.constants import SMALL_NS
from molgri.space.rotobj import SphereGridFactory, SphereGridNDim
from molgri.space.translations import TranslationParser
from molgri.naming import GridNameParser
from molgri.paths import PATH_OUTPUT_FULL_GRIDS
from molgri.space.utils import norm_per_axis, normalise_vectors, angle_between_vectors
from molgri.wrappers import save_or_use_saved, deprecated


class FullGrid:

    """
    A combination object that enables work a combination of three grids (provided by their names)

    Args:
        b_grid_name: body rotation grid (a 4D sphere grid of quaternions used to generate orientations)
        o_grid_name: origin rotation grid (a 3D sphere grid used to create approach vectors)
        t_grid_name: translation grid (a linear grid used to determine distances to origin)
    """

    def __init__(self, b_grid_name: str, o_grid_name: str, t_grid_name: str, use_saved: bool = True,
                 filter_non_unique: bool = False):

        """
        Args:
            b_grid_name: of the form 'ico_17'
            o_grid_name: of the form 'cube4D_12'
            t_grid_name: of the form '[1, 3, 4.5]'
            use_saved: try to obtain saved data if possible
            filter_non_unique: remove repeating points from spherical grids so that Voronoi cells can be built
        """
        b_grid_name = GridNameParser(b_grid_name, "b")
        self.b_rotations = SphereGridFactory.create(alg_name=b_grid_name.get_alg(), N=b_grid_name.get_N(),
                                                    dimensions=4, use_saved=use_saved,
                                                    filter_non_unique=filter_non_unique)
        o_grid_name = GridNameParser(o_grid_name, "o")
        self.o_rotations = SphereGridFactory.create(alg_name=o_grid_name.get_alg(), N=o_grid_name.get_N(),
                                                    dimensions=3, use_saved=use_saved,
                                                    filter_non_unique=filter_non_unique)
        self.o_positions = self.o_rotations.get_grid_as_array()
        self.t_grid = TranslationParser(t_grid_name)
        self.use_saved = use_saved

    def get_name(self) -> str:
        """Name that is appropriate for saving."""
        o_name = self.o_rotations.get_name(with_dim=False)
        b_name = self.b_rotations.get_name(with_dim=False)
        return f"o_{o_name}_b_{b_name}_t_{self.t_grid.grid_hash}"

    @save_or_use_saved
    def get_full_grid(self) -> Tuple[SphereGridNDim, SphereGridNDim, TranslationParser]:
        """Get all relevant sub-grids."""
        return self.b_rotations, self.o_rotations, self.t_grid

    def get_radii(self) -> NDArray:
        """
        Get the radii at which points are positioned. Result is in Angstroms.
        """
        return self.t_grid.get_trans_grid()

    def get_between_radii(self) -> NDArray:
        """
        Get the radii at which Voronoi cells of the position grid should be positioned. This should be right in-between
        two orientation point layers (except the first layer that is fully encapsulated by the first voronoi layer
        and the last one that is above the last one so that the last layer of points is right in-between the two last
        Voronoi cells

        Returns:
            an array of distances, same length as the self.get_radii array but with all distances larger than the
            corresponding point radii
        """
        radii = self.get_radii()

        # get increments to each radius, remove first one and add an extra one at the end with same distance as
        # second-to-last one
        increments = list(self.t_grid.get_increments())
        if len(increments) > 1:
            increments.pop(0)
            increments.append(increments[-1])
            increments = np.array(increments)
            increments = increments / 2
        else:
            increments = np.array(increments)

        between_radii = radii + increments
        return between_radii

    def get_body_rotations(self) -> Rotation:
        """Get a Rotation object (may encapsulate a list of rotations) from the body grid."""
        return Rotation.from_quat(self.b_rotations.get_grid_as_array())

    @save_or_use_saved
    def get_position_grid(self) -> NDArray:
        """
        Get a 'product' of o_grid and t_grid so you can visualise points in space at which COM of the second molecule
        will be positioned. Important: Those are points on spheres in real space.

        Returns:
            an array of shape (len_o_grid, len_t_grid, 3) in which the elements of result[0] have the first
            rotational position, each line at a new (increasing) distance, result[1] the next rotational position,
            again at all possible distances ...
        """
        dist_array = self.get_radii()
        o_grid = self.o_positions
        num_dist = len(dist_array)
        num_orient = len(o_grid)
        result = np.zeros((num_dist, num_orient, 3))
        for i, dist in enumerate(dist_array):
            result[i] = np.multiply(o_grid, dist)
            norms = norm_per_axis(result[i])
            assert np.allclose(norms, dist), "In a position grid, all vectors in i-th 'row' should have the same norm!"
        result = np.swapaxes(result, 0, 1)
        return result

    @save_or_use_saved
    def get_flat_position_grid(self) -> NDArray:
        """
        Get a position grid that is not structured layer-by-layer but is simply a 2D array of shape (N_t*N_o, 3) where
        N_t is the length of translation grid and N_o the length of orientation grid.
        """
        pos_grid = self.get_position_grid()
        pos_grid = np.swapaxes(pos_grid, 0, 1)
        return pos_grid.reshape((-1, 3))

    @deprecated
    def save_full_grid(self):
        """Save position grid."""
        np.save(f"{PATH_OUTPUT_FULL_GRIDS}position_grid_{self.get_name()}", self.get_position_grid())

    def get_full_voronoi_grid(self) -> Optional[FullVoronoiGrid]:
        """Get the corresponding FullVoronoiGrid object."""
        try:
            return FullVoronoiGrid(self)
        except AttributeError:
            return None

    def point2cell_position_grid(self, points_vector: NDArray) -> NDArray:
        """
        This method is used to back-map any points in 3D space to their corresponding Voronoi cell indices (of position
        grid). This means that, given an array of shape (k, 3) as points_vector, the result
        will be a vector of length k in which each item is an index of the flattened position grid. The point falls
        into the Voronoi cell associated with this index.
        """
        # determine index within a layer - the layer grid point to which the point vectors are closest
        rot_points = self.o_rotations.get_grid_as_array()
        # this automatically select the one of angles that is < pi
        angles = angle_between_vectors(points_vector, rot_points)
        indices_within_layer = np.argmin(angles, axis=1)

        # determine radii of cells
        norms = norm_per_axis(points_vector)
        layers = np.zeros((len(points_vector),))
        vor_radii = self.get_between_radii()

        # find the index of the layer to which each point belongs
        for i, norm in enumerate(norms):
            for j, vor_rad in enumerate(vor_radii):
                # because norm keeps the shape of the original array
                if norm[0] < vor_rad:
                    layers[i] = j
                    break
            else:
                layers[i] = np.NaN

        layer_len = len(rot_points)
        indices = layers * layer_len + indices_within_layer
        return indices

    def nan_free_assignments(self, points_vector: NDArray) -> Tuple[NDArray, NDArray]:
        """
        Same as point2cell_position_grid, but remove any cells that don't belong to the grid. In this way, there are no
        NaNs in the assignment array and it can be converted to the integer type
        """
        indices = self.point2cell_position_grid(points_vector)
        valid_points = points_vector[~np.isnan(indices)]
        indices = indices[~np.isnan(indices)]
        return valid_points, indices.astype(int)

    def points2cell_scipy(self, points_vector: NDArray) -> NDArray:
        """
        For comparison to self.point2cell_position_grid(), here is a method that assigns belonging to a grid point
        simply based on distance and not considering the spherical nature of the grid.
        """
        my_cdist = cdist(points_vector, self.get_flat_position_grid())
        return np.argmin(my_cdist, axis=1)


class FullVoronoiGrid:

    """
    Created from FullGrid, implements functionality that is specifically Voronoi-based.

    Enables calculations of all distances, areas, and volumes of/between cells.
    """

    def __init__(self, full_grid: FullGrid):
        self.full_grid = full_grid
        self.use_saved = self.full_grid.use_saved
        self.flat_positions = self.full_grid.get_flat_position_grid()
        self.all_sv = None
        self.get_voronoi_discretisation()

    ###################################################################################################################
    #                           basic creation/getter functions
    ###################################################################################################################

    def get_name(self) -> str:
        """Name for saving files."""
        return f"voronoi_{self.full_grid.get_name()}"

    def _change_voronoi_radius(self, sv: SphericalVoronoi, new_radius: float) -> SphericalVoronoi:
        """
        This is a helper function. Since a FullGrid consists of several layers of spheres in which the points are at
        exactly same places (just at different radii), it makes sense not to recalculate, but just to scale the radius,
        vertices and points out of which the SphericalVoronoi consists to a new radius.
        """
        sv.radius = new_radius
        sv.vertices = normalise_vectors(sv.vertices, length=new_radius)
        sv.points = normalise_vectors(sv.points, length=new_radius)
        # important that it's a copy!
        return copy(sv)

    @save_or_use_saved
    def get_voronoi_discretisation(self) -> list[SphericalVoronoi]:
        """
        Create a list of spherical voronoi-s that are identical except at different radii. The radii are set in such a
        way that there is always a Voronoi cell layer right in-between two layers of grid cells. (see FullGrid method
        get_between_radii for details.
        """
        if self.all_sv is None:
            unit_sph_voronoi = self.full_grid.o_rotations.get_spherical_voronoi_cells()
            between_radii = self.full_grid.get_between_radii()
            self.all_sv = [self._change_voronoi_radius(unit_sph_voronoi, r) for r in between_radii]
        return self.all_sv

    ###################################################################################################################
    #                           point helper functions
    ###################################################################################################################

    def _at_same_radius(self, point1: Point, point2: Point) -> bool:
        """Check that two points belong to the same layer"""
        return bool(np.isclose(point1.d_to_origin, point2.d_to_origin))

    def _are_neighbours(self, point1: Point, point2: Point) -> bool:
        """
        Check whether points are neighbours in any way. We define neighbours as those sharing surface and NOT sharing
        only vertex points or lines.
        """
        index1 = point1.index_position_grid
        index2 = point2.index_position_grid
        return self.get_division_area(index1, index2, print_message=False) is not None

    def _are_sideways_neighbours(self, point1: Point, point2: Point) -> bool:
        """
        Check whether points belong to the same radius and are neighbours.
        """
        return self._at_same_radius(point1, point2) and self._are_neighbours(point1, point2)

    def _are_on_same_ray(self, point1: Point, point2: Point) -> bool:
        """
        Two points are on the same ray if they are on the same vector from origin (may be at different distances)
        """
        normalised1 = point1.get_normalised_point()
        normalised2 = point2.get_normalised_point()
        return np.allclose(normalised1, normalised2)

    def _point1_right_above_point2(self, point1: Point, point2: Point) -> bool:
        """
        Right above should be interpreted spherically.

        Conditions: on the same ray + radius of point1 one unit bigger
        """
        radial_index1 = point1.index_radial
        radial_index2 = point2.index_radial
        return self._are_on_same_ray(point1, point2) and radial_index1 == radial_index2 + 1

    def _point2_right_above_point1(self, point1: Point, point2: Point) -> bool:
        """See _point1_right_above_point2."""
        radial_index1 = point1.index_radial
        radial_index2 = point2.index_radial
        return self._are_on_same_ray(point1, point2) and radial_index1 + 1 == radial_index2

    ###################################################################################################################
    #                           useful properties
    ###################################################################################################################

    def find_voronoi_vertices_of_point(self, point_index: int, which: str = "all") -> NDArray:
        """
        Using an index (from flattened position grid), find which voronoi vertices belong to this point.

        Args:
            point_index: for which point in flattened position grid the vertices should be found
            which: which vertices: all, upper or lower

        Returns:
            an array of vertices, each row is a 3D point.
        """
        my_point = Point(point_index, self)

        if which == "all":
            vertices = my_point.get_vertices()
        elif which == "upper":
            vertices = my_point.get_vertices_above()
        elif which == "lower":
            vertices = my_point.get_vertices_below()
        else:
            raise ValueError("The value of which not recognised, select 'all', 'upper', 'lower'.")
        return vertices

    def get_distance_between_centers(self, index_1: int, index_2: int, print_message=True) -> Optional[float]:
        """
        Calculate the distance between two position grid points selected by their indices. Optionally print message
        if they are not neighbours.

        There are three options:
            - point1 is right above point2 or vide versa -> the distance is measured in a straight line from the center
            - point1 and point2 are sideways neighbours -> the distance is measured on the circumference of their radius
            - point1 and point2 are not neighbours -> return None

        Returns:
            None if not neighbours, distance in angstrom else
        """
        point_1 = Point(index_1, self)
        point_2 = Point(index_2, self)

        if self._point1_right_above_point2(point_1, point_2) or self._point2_right_above_point1(point_1, point_2):
            return np.abs(point_1.d_to_origin - point_2.d_to_origin)
        elif self._are_sideways_neighbours(point_1, point_2):
            radius = point_1.d_to_origin
            theta = angle_between_vectors(point_1.point, point_2.point)
            # length of arc
            return theta * radius
        else:
            if print_message:
                print(f"Points {index_1} and {index_2} are not neighbours.")
            return None

    def get_division_area(self, index_1: int, index_2: int, print_message: bool = True) -> Optional[float]:
        """
        Calculate the area (in Angstrom squared) that is the border area between two Voronoi cells. This is either
        a curved area (a part of a sphere) if the two cells are one above the other or a flat, part of circle or
        circular ring if the cells are neighbours at the same level. If points are not neighbours, returns None.
        """
        point_1 = Point(index_1, self)
        point_2 = Point(index_2, self)

        # if they are sideways neighbours
        if self._at_same_radius(point_1, point_2):
            # vertices_above
            vertices_1a = point_1.get_vertices_above()
            vertices_2a = point_2.get_vertices_above()
            r_larger = np.linalg.norm(vertices_1a[0])
            set_vertices_1a = set([tuple(v) for v in vertices_1a])
            set_vertices_2a = set([tuple(v) for v in vertices_2a])
            # vertices that are above point 1 and point 2
            intersection_a = set_vertices_1a.intersection(set_vertices_2a)
            # vertices below - only important to determine radius
            vertices_1b = point_1.get_vertices_below()
            r_smaller = np.linalg.norm(vertices_1b[0])
            if len(intersection_a) != 2:
                if print_message:
                    print(f"Points {index_1} and {index_2} are not neighbours.")
                return None
            else:
                # angle will be determined by the vector from origin to both points above
                intersection_list = list(intersection_a)
                theta = angle_between_vectors(np.array(intersection_list[0]), np.array(intersection_list[1]))
                return theta / 2 * (r_larger ** 2 - r_smaller ** 2)
        # if point_1 right above point_2
        if self._point1_right_above_point2(point_1, point_2):
            return point_2.get_area_above()
        if self._point2_right_above_point1(point_1, point_2):
            return point_1.get_area_above()
        # if no exit point so far
        if print_message:
            print(f"Points {index_1} and {index_2} are not neighbours.")
        return None

    def get_volume(self, index: int) -> float:
        """
        Get the volume of any cell in FullVoronoiGrid, defined by its index in flattened position grid.
        """
        point = Point(index, self)
        return point.get_cell_volume()

    @save_or_use_saved
    def get_all_voronoi_volumes(self) -> NDArray:
        """
        Get an array in the same order as flat position grid, listing the volumes of Voronoi cells.
        """
        N = len(self.flat_positions)
        volumes = np.zeros((N,))
        for i in range(0, N):
            volumes[i] = self.get_volume(i)
        return volumes

    def _get_property_all_pairs(self, method: Callable) -> csc_array:
        """
        Helper method for any property that is dependent on a pair of indices. Examples: obtaining all distances
        between cell centers or all areas between cells. Always symmetrical - value at (i, j) equals the one at (j, i)

        Args:
            method: must have a signature (index1: int, index2: int, print_message: boolean)

        Returns:
            a sparse matrix of shape (len_flat_position_array, len_flat_position_array)
        """
        data = []
        row_indices = []
        column_indices = []
        N_pos_array = len(self.flat_positions)
        for i in range(N_pos_array):
            for j in range(i + 1, N_pos_array):
                my_property = method(i, j, print_message=False)
                if my_property is not None:
                    # this value will be added to coordinates (i, j) and (j, i)
                    data.extend([my_property, my_property])
                    row_indices.extend([i, j])
                    column_indices.extend([j, i])
        sparse_property = coo_array((data, (row_indices, column_indices)), shape=(N_pos_array, N_pos_array))
        return sparse_property.tocsc()

    @save_or_use_saved
    def get_all_voronoi_surfaces(self) -> csc_array:
        """
        If l is the length of the flattened position array, returns a lxl (sparse) array where the i-th row and j-th
        column (as well as the j-th row and the i-th column) represent the size of the Voronoi surface between points
        i and j in position grid. If the points do not share a division area, no value will be set.
        """
        return self._get_property_all_pairs(self.get_division_area)

    @save_or_use_saved
    def get_all_distances_between_centers(self) -> csc_array:
        """
        Get a sparse matrix where for all sets of neighbouring cells the distance between Voronoi centers is provided.
        Therefore, the value at [i][j] equals the value at [j][i]. For more information, check
        self.get_distance_between_centers.
        """
        return self._get_property_all_pairs(self.get_distance_between_centers)

    def get_all_voronoi_surfaces_as_numpy(self) -> NDArray:
        """See self.get_all_voronoi_surfaces, only transforms sparse array to normal array."""
        return self.get_all_voronoi_surfaces().toarray(order='C')

    def get_all_distances_between_centers_as_numpy(self) -> NDArray:
        """See self.get_all_distances_between_centers, only transforms sparse array to normal array."""
        return self.get_all_distances_between_centers().toarray(order='C')


class Point:

    """
    A Point represents a single cell in a particular FullVoronoiGrid. It holds all relevant information, eg. the
    index of the cell within a single layer and in a full flattened position grid. It enables the identification
    of Voronoi vertices and connected calculations (distances, areas, volumes).
    """

    def __init__(self, index_position_grid: int, full_sv: FullVoronoiGrid):
        self.full_sv = full_sv
        self.index_position_grid: int = index_position_grid
        self.point: NDArray = self.full_sv.flat_positions[index_position_grid]
        self.d_to_origin: float = np.linalg.norm(self.point)
        self.index_radial: int = self._find_index_radial()
        self.index_within_sphere: int = self._find_index_within_sphere()

    def get_normalised_point(self) -> NDArray:
        """Get the vector to the grid point (center of Voronoi cell) normalised to length 1."""
        return normalise_vectors(self.point, length=1)

    def _find_index_radial(self) -> int:
        """Find to which radial layer of points this point belongs."""
        point_radii = self.full_sv.full_grid.get_radii()
        for i, dist in enumerate(point_radii):
            if np.isclose(dist, self.d_to_origin):
                return i
        else:
            raise ValueError("The norm of the point not close to any of the radii.")

    def _find_index_within_sphere(self) -> int:
        """Find the index of the point within a single layer of possible orientations."""
        radial_index = self.index_radial
        num_o_rot = len(self.full_sv.full_grid.o_rotations.get_grid_as_array())
        return self.index_position_grid - num_o_rot * radial_index

    def _find_index_sv_above(self) -> Optional[int]:
        for i, sv in enumerate(self.full_sv.get_voronoi_discretisation()):
            if sv.radius > self.d_to_origin:
                return i
        else:
            # the point is outside the largest voronoi sphere
            return None

    def _get_sv_above(self) -> SphericalVoronoi:
        """Get the spherical Voronoi with the first radius that is larger than point radius."""
        return self.full_sv.get_voronoi_discretisation()[self._find_index_sv_above()]

    def _get_sv_below(self) -> Optional[SphericalVoronoi]:
        """Get the spherical Voronoi with the largest radius that is smaller than point radius. If the point is in the
        first layer, return None."""
        index_above = self._find_index_sv_above()
        if index_above != 0:
            return self.full_sv.get_voronoi_discretisation()[index_above-1]
        else:
            return None

    ##################################################################################################################
    #                            GETTERS - DISTANCES, AREAS, VOLUMES
    ##################################################################################################################

    def get_radius_above(self) -> float:
        """Get the radius of the SphericalVoronoi cell that is the upper surface of the cell."""
        sv_above = self._get_sv_above()
        return sv_above.radius

    def get_radius_below(self) -> float:
        """Get the radius of the SphericalVoronoi cell that is the lower surface of the cell (return zero if there is
        no Voronoi layer below)."""
        sv_below = self._get_sv_below()
        if sv_below is None:
            return 0.0
        else:
            return sv_below.radius

    def get_area_above(self) -> float:
        """Get the area of the Voronoi surface that is the upper side of the cell (curved surface, part of sphere)."""
        sv_above = self._get_sv_above()
        areas = sv_above.calculate_areas()
        return areas[self.index_within_sphere]

    def get_area_below(self) -> float:
        """Get the area of the Voronoi surface that is the lower side of the cell (curved surface, part of sphere)."""
        sv_below = self._get_sv_below()
        if sv_below is None:
            return 0.0
        else:
            areas = sv_below.calculate_areas()
            return areas[self.index_within_sphere]

    def get_vertices_above(self) -> NDArray:
        """Get the vertices of this cell that belong to the SphericalVoronoi above the point."""
        sv_above = self._get_sv_above()
        regions = sv_above.regions[self.index_within_sphere]
        vertices_above = sv_above.vertices[regions]
        return vertices_above

    def get_vertices_below(self) -> NDArray:
        """Get the vertices of this cell that belong to the SphericalVoronoi below the point (or just the origin
        if the point belongs to the first layer."""
        sv_below = self._get_sv_below()
        if sv_below is None:
            vertices_below = np.zeros((1, 3))
        else:
            regions = sv_below.regions[self.index_within_sphere]
            vertices_below = sv_below.vertices[regions]

        return vertices_below

    def get_vertices(self) -> NDArray:
        """Get all vertices of this cell as a single array."""
        vertices_above = self.get_vertices_above()
        vertices_below = self.get_vertices_below()

        return np.concatenate((vertices_above, vertices_below))

    def get_cell_volume(self) -> float:
        """Get the volume of this cell (calculated as the difference between the part of sphere volume at radius above
        minus the same area at smaller radius)."""
        radius_above = self.get_radius_above()
        radius_below = self.get_radius_below()
        area_above = self.get_area_above()
        area_below = self.get_area_below()
        volume = 1/3 * (radius_above * area_above - radius_below * area_below)
        return volume


class ConvergenceFullGridO:
    
    """
    This object is used to study the convergence of properties as the number of points in the o_grid of FullGrid 
    changes. Mostly used to plot how voronoi volumes, calculation times ... converge.
    """

    def __init__(self, b_grid_name: str, t_grid_name: str,  o_alg_name: str, N_set: tuple = None,
                 use_saved: bool = False, **kwargs):
        if N_set is None:
            N_set = SMALL_NS
        self.N_set = N_set
        self.alg_name = o_alg_name
        self.use_saved = use_saved
        self.list_full_grids = self.create(b_grid_name=b_grid_name, t_grid_name=t_grid_name,  o_alg_name=o_alg_name,
                                           N_set=self.N_set, use_saved=use_saved, **kwargs)

    def get_name(self) -> str:
        """Get name for saving."""
        b_name = self.list_full_grids[0].b_rotations.get_name(with_dim=False)
        t_name = self.list_full_grids[0].t_grid.grid_hash
        return f"convergence_o_{self.alg_name}_b_{b_name}_t_{t_name}"

    @classmethod
    def create(cls,  b_grid_name: str, t_grid_name: str,  o_alg_name: str, N_set: tuple, **kwargs) -> list:
        """
        Build the FullGrid object for each of the given Ns in N_set. This is the create function that is only
        called by __init__.
        """
        list_full_grids = []
        for N in N_set:
            fg = FullGrid(b_grid_name=b_grid_name, o_grid_name=f"{o_alg_name}_{N}", t_grid_name=t_grid_name, **kwargs)
            list_full_grids.append(fg)
        return list_full_grids

    @save_or_use_saved
    def get_voronoi_volumes(self) -> pd.DataFrame:
        """
        Get all volumes of cells for all choices of self.N_set in order to create convergence plots.
        """
        data = []
        for N, fg in zip(self.N_set, self.list_full_grids):
            N_real = fg.o_rotations.get_N()
            vor_radius = list(fg.get_between_radii())
            vor_radius.insert(0, 0)
            vor_radius = np.array(vor_radius)
            fvg = fg.get_full_voronoi_grid()
            ideal_volumes = 4/3 * pi * (vor_radius[1:]**3 - vor_radius[:-1]**3) / N_real
            real_volumes = fvg.get_all_voronoi_volumes()
            for i, volume in enumerate(real_volumes):
                layer = i//N_real
                data.append([N_real, layer, ideal_volumes[i//N_real], volume])
        df = pd.DataFrame(data, columns=["N", "layer", "ideal volume", "Voronoi cell volume"])
        return df



"""
The molgri.plotting sub-package builds on matplotlib and seaborn to provide different visualisations of data based on
a central object that should be represented. For example, SphereGridPlot is initiated with a SphereGridNDim object
and provides various methods to plot and animate the data contained in this object.

All plotting implementations extend basic set-up and saving options set up in the abstract module.
"""


"""
All boolean/assertion helper functions.

This module implements helper functions that can be used by different test/other functions. The determining
characteristics of all functions in this module is that they operate with booleans (either return a boolean
answer or assert that a condition is true.

Example functions that belong to this module:
 - check if two quaternion sets represent the same rotation
 - check if all norms of vectors in the array are similar
 - check if given points form a square/a cube ...
"""

import numpy as np
from numpy.typing import NDArray
from scipy.constants import pi
from scipy.spatial import distance_matrix
from scipy.spatial.distance import cdist
from scipy.spatial.transform import Rotation

from molgri.space.utils import norm_per_axis
from molgri.constants import UNIQUE_TOL

##########################################################################################################
#                                          SPACE TOOLS                                                   #
##########################################################################################################


def check_equality(arr1: NDArray, arr2: NDArray, atol: float = None, rtol: float = None) -> bool:
    """
    Use the numpy function np.allclose to compare two arrays and return True if they are all equal. This function
    is a wrapper where I can set my preferred absolute and relative tolerance
    """
    if atol is None:
        atol = 1e-8
    if rtol is None:
        rtol = 1e-5
    return np.allclose(arr1, arr2, atol=atol, rtol=rtol)


def all_row_norms_similar(my_array: NDArray, atol: float = None, rtol: float = None) -> NDArray:
    """
    Assert that in an 2D array each row has the same norm (up to the floating point tolerance).

    Returns:
        the array of norms in the same shape as my_array
    """
    is_array_with_d_dim_r_rows_c_columns(my_array, d=2)
    axis = 1
    all_norms = norm_per_axis(my_array, axis=axis)
    average_norm = np.average(all_norms)
    assert check_equality(all_norms, average_norm, atol=atol, rtol=rtol), "The norms of all rows are not equal"
    return all_norms


def all_row_norms_equal_k(my_array: NDArray, k: float, atol: float = None, rtol: float = None) -> NDArray:
    """
    Same as all_row_norms_similar, but also test that the norm equals k.
    """
    my_norms = all_row_norms_similar(my_array=my_array, atol=atol, rtol=rtol)
    assert check_equality(my_norms, np.array(k), atol=atol, rtol=rtol), "The norms are not equal to k"
    return my_norms


def is_array_with_d_dim_r_rows_c_columns(my_array: NDArray, d: int = None, r: int = None, c: int = None):
    """
    Assert that the object is an array. If you specify d, r, c, it will check if this number of dimensions, rows, and/or
    columns are present.
    """
    assert type(my_array) == np.ndarray, "The first argument is not an array"
    # only check if dimension if d specified
    if d is not None:
        assert len(my_array.shape) == d, f"The dimension of an array is not d: {len(my_array.shape)}=!={d}"
    if r is not None:
        assert my_array.shape[0] == r, f"The number of rows is not r: {my_array.shape[0]}=!={r}"
    if c is not None:
        assert my_array.shape[1] == c, f"The number of columns is not c: {my_array.shape[1]}=!={c}"


def all_rows_unique(my_array: NDArray, tol: int = UNIQUE_TOL):
    """
    Check if all rows of the array are unique up to tol number of decimal places.
    """
    my_unique = np.unique(my_array.round(tol), axis=0)
    difference = np.abs(len(my_array) - len(my_unique))
    assert len(my_array) == len(my_unique), f"{difference} elements of an array are not unique up to tolerance."


def form_square_array(my_array: NDArray, dec_places=7) -> bool:
    """
    From an array of exactly 4 points, determine if they form a square (can be in space and not on a 2D plane).

    Args:
        my_array: array of shape (4, d) where d is the number of dimensions.
        dec_places: to how many decimal places to round when determining uniqueness

    Returns:
        True if points form a square, else False
    """
    distances = cdist(my_array, my_array)
    return _form_square(my_array, distances, dec_places=dec_places)


def _form_square(points, distances, dec_places=7) -> bool:
    """
    Given a set of points and already known distances between all pairs, determine if points form a square.
    """
    is_array_with_d_dim_r_rows_c_columns(points, d=2, r=4)
    # in each row of distances, the values must be: 1x0, 2xa (side length), 1xnp.sqrt(2)*a (diagonal length)
    dists, counts = np.unique(np.round(distances[0], dec_places), return_counts=True)   # sorts smallest to largest
    # from the first row, determine a and d = np.sqrt(2)*a
    try:
        dist_a = dists[1]
        dist_d = np.sqrt(2) * dist_a
    # IndexErrors occur when not a correct number of unique values and indicate that the point array isn't a square
    except IndexError:
        return False
    # now check this for every row
    for i, row in enumerate(distances):
        dists, counts = np.unique(np.round(row, dec_places), return_counts=True)  # sorts smallest to largest
        try:
            once_dist_zero = np.isclose(dists[0], 0) and counts[0] == 1
            twice_dist_a = counts[1] == 2
            once_dist_d = counts[2] == 1 and np.isclose(dists[2], dist_d)
        except IndexError:
            return False
        # if any of the conditions do not apply, return False immediately
        if not (once_dist_d and twice_dist_a and once_dist_zero):
            return False
    return True


def form_cube(my_array: NDArray, test_angles=False) -> bool:
    """
    Similarly to form_square, check if the 8 points supplied as rows in an array form a cube. The points may have
    >= 3 dimensions.

    Args:
        my_array: 2-dimensional array with 8 rows and at least 3 columns. It should be checked if the points form a
                  3D cube
        test_angles: select True if you want a more precise test that also looks at angles

    Returns:
        True if points form a cube, else False
    """
    is_array_with_d_dim_r_rows_c_columns(my_array, d=2, r=8)
    assert my_array.shape[1] >= 3, f"Only {my_array.shape[1]} columns given. " \
                                   f"Points with <3 dimensions cannot possibly form a cube!"
    # idea source: https://math.stackexchange.com/questions/1629899/given-eight-vertices-how-to-verify-they-form-a-cube
    distances = distance_matrix(my_array, my_array)
    # side length is the shortest distance occurring in distance matrix (except for 0
    a = np.min(distances[np.nonzero(distances)])
    # 24 elements should be equal to a
    if not np.count_nonzero(np.isclose(distances, a)) == 24:
        return False
    # 24 elements are face diagonals
    if not np.count_nonzero(np.isclose(distances, a*np.sqrt(2))) == 24:
        return False
    # 8 elements should be volume diagonals
    if not np.count_nonzero(np.isclose(distances, a * np.sqrt(3))) == 8:
        return False
    # the side length is a and as a consequence, distances to other points should be: 1x0, 3xa, 3xsqrt(2)*a, 1xsqrt(3)*a
    for row in distances:
        if not np.allclose(sorted(row), sorted([0, a, a, a, a*np.sqrt(2), a*np.sqrt(2), a*np.sqrt(2), a*np.sqrt(3)])):
            return False
    # only selected angles possible
    if test_angles:
        tetrahedron_a1 = np.arccos(-1 / 3)
        tetrahedron_a2 = np.arccos(1 / 3)
        all_angle_choices = [0, np.pi/2, np.pi, np.pi/3, tetrahedron_a1, tetrahedron_a2]
        for vec1 in my_array:
            for vec2 in my_array:
                angle_points = np.arccos(np.clip(np.dot(vec1, vec2), -1.0, 1.0))
                if not np.any(np.isclose(angle_points, all_angle_choices, atol=1e-7)):
                    return False
    # finally, if all tests right
    return True


def quaternion_in_array(quat: NDArray, quat_array: NDArray) -> bool:
    """
    Check if a quaternion q or its equivalent complement -q is present in the quaternion array quat_array.
    """
    quat1 = quat[np.newaxis, :]
    for quat2 in quat_array:
        if two_sets_of_quaternions_equal(quat1, quat2[np.newaxis, :]):
            return True
    return False


def two_sets_of_quaternions_equal(quat1: NDArray, quat2: NDArray) -> bool:
    """
    This test is necessary because for quaternions, q and -q represent the same rotation. You therefore cannot simply
    use np.allclose to check if two sets of rotations represented with quaternions are the same. This function checks
    if all rows of two arrays are the same up to a flipped sign.
    """
    assert quat1.shape == quat2.shape
    assert quat1.shape[1] == 4
    # quaternions are the same if they are equal up to a +- sign
    # I have checked this fact and it is mathematically correct
    for q1, q2 in zip(quat1, quat2):
        if not (np.allclose(q1, q2) or np.allclose(q1, -q2)):
            return False
    return True


def assert_two_sets_of_eulers_equal(euler1: NDArray, euler2: NDArray):
    # see: https://en.wikipedia.org/wiki/Euler_angles#Signs,_ranges_and_conventions
    assert euler1.shape == euler2.shape
    assert euler1.shape[1] == 3
    # first of all, check that the corresponding rot matrices are the same
    rot1 = Rotation.from_euler("ZYX", euler1)
    rot2 = Rotation.from_euler("ZYX", euler2)
    assert np.allclose(rot1.as_matrix(), rot2.as_matrix())

    # further checks
    for e1, e2 in zip(euler1, euler2):
        # if within [-pi/2, pi/2], euler angles must match exactly
        for comp1, comp2 in zip(e1, e2):
            if -pi/2 <= comp1 <= pi/2 and -pi/2 <= comp2 <= pi/2:
                assert np.isclose(comp1, comp2)
        # first and last angle must match up to pi
        assert np.isclose(e1[0] % pi, e2[0] % pi)
        assert np.isclose(e1[-1] % pi, e2[-1] % pi)


"""
Plots of ITS, eigenvectors and eigenvalues of transition/rate matrix.

A collection of methods to visualise the SqRA or MSM objects.
"""

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from molgri.plotting.abstract import RepresentationCollection
from molgri.molecules.transitions import TransitionModel
from molgri.plotting.fullgrid_plots import FullGridPlot
from molgri.constants import DIM_SQUARE


class TransitionPlot(RepresentationCollection):

    def __init__(self, transition_obj: TransitionModel, *args, **kwargs):
        self.transition_obj = transition_obj
        data_name = self.transition_obj.get_name()
        super().__init__(data_name, *args, **kwargs)

    def make_its_plot(self, fig=None, ax=None, save=True, num_eigenv=6, as_line=False):
        """
        Plot iterative timescales.
        """
        self._create_fig_ax(ax=ax, fig=fig)
        tau_array = self.transition_obj.tau_array
        eigenvals, eigenvecs = self.transition_obj.get_eigenval_eigenvec(num_eigenv=num_eigenv)
        eigenvals = np.array(eigenvals)
        dt = self.transition_obj.sim_hist.parsed_trajectory.dt

        if not as_line:
            for j in range(1, num_eigenv):
                to_plot_abs = np.array(-tau_array * dt / np.log(np.abs(eigenvals[:, j])))
                sns.lineplot(x=tau_array * dt, y=to_plot_abs,
                             ax=ax, legend=False)
        else:
            # for SQRA plot vertical lines
            eigenvals = eigenvals[0]
            for j in range(1, num_eigenv):
                x_min, x_max = self.ax.get_xlim()
                min_value = np.min([0, x_min])
                max_value = np.max([1, x_max])
                self.ax.hlines(- 1 / eigenvals[j], min_value, max_value, color="black", ls="--")

        self.ax.set_xlim(left=0, right=tau_array[-1] * dt)
        self.ax.set_xlabel(r"$\tau$")
        self.ax.set_ylabel(r"ITS")
        tau_array_with_zero = (tau_array * dt).tolist()
        tau_array_with_zero.append(0)
        tau_array_with_zero.sort()
        self.ax.fill_between(tau_array_with_zero, tau_array_with_zero, color="grey", alpha=0.5)

        if save:
            self._save_plot_type(f"its")

    def make_eigenvalues_plot(self, fig=None, ax=None, save=True, num_eigenv=None):
        """
        Visualize the eigenvalues of rate matrix.
        """

        self._create_fig_ax(ax=ax, fig=fig)

        eigenvals, _ = self.transition_obj.get_eigenval_eigenvec()
        eigenvals = np.array(eigenvals)[0]

        if num_eigenv:
            eigenvals = eigenvals[:num_eigenv]

        xs = np.linspace(0, 1, num=len(eigenvals))
        self.ax.scatter(xs, eigenvals, s=5, c="black")
        for i, eigenw in enumerate(eigenvals):
            self.ax.vlines(xs[i], eigenw, 0, linewidth=0.5, color="black")
        self.ax.hlines(0, 0, 1, color="black")
        self.ax.set_ylabel(f"Eigenvalues")
        self.ax.axes.get_xaxis().set_visible(False)

        if save:
            self._save_plot_type(f"eigenvalues")

    def make_eigenvectors_plot(self, ax=None, fig=None, save=True, num_eigenv: int = 5, projection="3d"):
        """
        Visualize the energy surface and the first num (default=3) eigenvectors
        """

        self.fig, self.ax = plt.subplots(1, num_eigenv, subplot_kw={"projection": projection},
                                         figsize=(DIM_SQUARE[0], num_eigenv*DIM_SQUARE[0]))

        for i, subax in enumerate(self.ax.ravel()):
            self.make_one_eigenvector_plot(i, ax=subax, fig=self.fig, projection=projection, save=False)

        if save:
            self._save_plot_type(f"eigenvectors_{projection}")

    def make_one_eigenvector_plot(self, eigenvec_index: int, ax=None, fig=None, save=True, projection="3d",
                                  animate_rot=False):
        self._create_fig_ax(ax=ax, fig=fig, projection=projection)

        eigenvals, eigenvecs = self.transition_obj.get_eigenval_eigenvec()

        # shape: (number_taus, number_cells, num_eigenvectors)
        eigenvecs = eigenvecs[0]  # values for the first tau
        eigenvecs = eigenvecs.T

        fgp = FullGridPlot(self.transition_obj.sim_hist.full_grid, default_complexity_level="half_empty")
        if projection == "3d":
            fgp.make_full_voronoi_plot(ax=self.ax, fig=self.fig, plot_vertex_points=False, save=False)
        fgp.make_position_plot(ax=self.ax, fig=self.fig, save=False, c=eigenvecs[eigenvec_index], animate_rot=False,
                               projection=projection)
        self.ax.set_title(f"Eigenv. {eigenvec_index}")
        if save:
            self._save_plot_type(f"eigenvector_{eigenvec_index}_{projection}")
        if animate_rot:
            return self._animate_figure_view(self.fig, self.ax, f"eigenvector_{eigenvec_index}_rotated")


if __name__ == "__main__":
    if __name__ == "__main__":
        from molgri.molecules._load_examples import load_molgri_data, load_simulation_data
        from molgri.space.fullgrid import FullGrid
        from molgri.molecules.transitions import MSM, SQRA, SimulationHistogram
        from molgri.plotting.other_plots import ArrayPlot
        from molgri.plotting.molecule_plots import TrajectoryPlot

        USE_SAVED = False

        # TRANSITION MATRIX

        # parsed_sim = load_simulation_data()
        # # define some full grid to assign to
        # full_grid = FullGrid(t_grid_name="linspace(3, 13, 4)", o_grid_name="ico_20", b_grid_name="zero")
        #
        # combined_sim = SimulationHistogram(parsed_sim, full_grid)
        # msm = MSM(combined_sim, energy_type="Potential Energy", use_saved=USE_SAVED)
        # transition_matrix = msm.get_transitions_matrix(
        #
        # )
        # ArrayPlot(transition_matrix[0], default_context="talk").make_heatmap_plot(save=True)
        # tp_msm = TransitionPlot(msm, default_context="talk")
        # tp_msm.make_its_plot(save=True)
        # tp_msm.make_eigenvectors_plot(num_eigenv=3)
        # tp_msm.make_eigenvalues_plot()
        #
        # # RATE MATRIX
        #
        molgri_pt = load_molgri_data()

        tp = TrajectoryPlot(molgri_pt)

        full_grid_m = FullGrid(t_grid_name="linspace(0.8, 1.5, 10)", o_grid_name="ico_50", b_grid_name="zero")
        #
        combined_molgri = SimulationHistogram(molgri_pt, full_grid_m)
        sqra = SQRA(combined_molgri, energy_type="Potential Energy", use_saved=USE_SAVED)
        #rates_matrix = sqra.get_transitions_matrix()
        #ArrayPlot(rates_matrix[0], default_context="talk").make_heatmap_plot(save=True)

        tp_sqra = TransitionPlot(sqra, default_context="talk")
        tp_sqra.make_its_plot(save=True, as_line=True)
        tp_sqra.make_eigenvectors_plot(num_eigenv=3)
        tp_sqra.make_eigenvalues_plot()

        # full_grid = FullGrid(t_grid_name="[5, 15]", o_grid_name="ico_10", b_grid_name="zero")
        # tp = TrajectoryPlot(parsed_sim)
        # ani = tp.make_COM_plot(animate_rot=True, projection="3d", save=True, fg=full_grid)


"""
Naming conventions are defined here.
"""

from molgri.constants import GRID_ALGORITHMS, ZERO_ALGORITHM, DEFAULT_ALGORITHM_O, DEFAULT_ALGORITHM_B


class NameParser:

    def __init__(self, name_string: str):
        self.name_string = name_string
        self.N = self._find_a_number()
        self.algo = self._find_algorithm()
        self.dim = self._find_dimensions()

    def _find_a_number(self) -> int or None:
        """
        Try to find an integer representing number of grid points anywhere in the name.

        Returns:
            the number of points as an integer, if it can be found, else None

        Raises:
            ValueError if more than one integer present in the string (e.g. 'ico_12_17')
        """
        split_string = self.name_string.split("_")
        candidates = []
        for fragment in split_string:
            if fragment.isnumeric():
                candidates.append(int(fragment))
        # >= 2 numbers found in the string
        if len(candidates) > 1:
            raise ValueError(f"Found two or more numbers in grid name {self.name_string},"
                             f" can't determine num of points.")
        # exactly one number in the string -> return it
        elif len(candidates) == 1:
            return candidates[0]
        # no number in the string -> return None
        else:
            return None

    def _find_algorithm(self) -> str or None:
        split_string = self.name_string.split("_")
        candidates = []
        for fragment in split_string:
            if fragment in GRID_ALGORITHMS:
                candidates.append(fragment)
            elif fragment.lower() == "none":
                candidates.append(ZERO_ALGORITHM)
        # >= 2 algorithms found in the string
        if len(candidates) > 1:
            raise ValueError(f"Found two or more algorithm names in grid name {self.name_string}, can't decide.")
        # exactly one algorithm in the string -> return it
        elif len(candidates) == 1:
            return candidates[0]
        # no algorithm given -> None
        else:
            return None

    def _find_dimensions(self) -> str or None:
        split_string = self.name_string.split("_")
        candidates = []
        for fragment in split_string:
            if len(fragment) == 2 and fragment[-1] == "d" and fragment[0].isnumeric():
                candidates.append(int(fragment))
        if len(candidates) > 1:
            raise ValueError(f"Found two or more dimension names in grid name {self.name_string}, can't decide.")
        # exactly one algorithm in the string -> return it
        elif len(candidates) == 1:
            return candidates[0]
        # no algorithm given -> None
        else:
            return None


class FullGridNameParser:

    def __init__(self, name_string: str):
        split_name = name_string.split("_")
        self.b_grid_name = None
        self.o_grid_name = None
        self.t_grid_name = None
        for i, split_part in enumerate(split_name):
            if split_part == "b":
                try:
                    self.b_grid_name = GridNameParser(
                        f"{split_name[i + 1]}_{split_name[i + 2]}", "b").get_standard_grid_name()
                except IndexError:
                    self.b_grid_name = GridNameParser(f"{split_name[i + 1]}", "o").get_standard_grid_name()
            elif split_part == "o":
                try:
                    self.o_grid_name = GridNameParser(
                        f"{split_name[i + 1]}_{split_name[i + 2]}").get_standard_grid_name()
                except IndexError:
                    self.o_grid_name = GridNameParser(f"{split_name[i + 1]}").get_standard_grid_name()
            elif split_part == "t":
                self.t_grid_name = f"{split_name[i + 1]}"

    def get_standard_full_grid_name(self):
        return f"o_{self.o_grid_name}_b_{self.b_grid_name}_t_{self.t_grid_name}"

    def get_num_b_rot(self):
        return int(self.b_grid_name.split("_")[1])

    def get_num_o_rot(self):
        return int(self.o_grid_name.split("_")[1])


class GridNameParser(NameParser):
    """
    Differently than pure NameParser, GridNameParser raises errors if the name doesn't correspond to a standard grid
    name.
    """
    # TODO: don't simply pass if incorrectly written alg name!
    def __init__(self, name_string: str, o_or_b="o"):
        super().__init__(name_string)
        # num of points 0 or 1 -> always zero algorithm; selected zero algorithm -> always num of points is 1
        if (self.N in (1, 0)) or (self.N is None and self.algo == ZERO_ALGORITHM):
            self.algo = ZERO_ALGORITHM
            self.N = 1
        # ERROR - zero algorithm but a larger number of points provided
        elif self.algo == ZERO_ALGORITHM and self.N > 1:
            raise ValueError(f"Zero algorithm selected but number of points {self.N}>1 in {self.name_string}.")
        # ERROR - no points but also not zero algorithm
        elif self.N is None and (self.algo != ZERO_ALGORITHM and self.algo is not None):
            raise ValueError(f"An algorithm provided ({self.algo}) but not number of points in {self.name_string}")
        # algorithm not provided but > 1 points -> default algorithm
        elif self.algo is None and self.N > 1 and o_or_b == "o":
            self.algo = DEFAULT_ALGORITHM_O
        elif self.algo is None and self.N > 1 and o_or_b == "b":
            self.algo = DEFAULT_ALGORITHM_B
        # ERROR - absolutely nothing provided
        elif self.algo is None and self.N is None:
            raise ValueError(f"Algorithm name and number of grid points not recognised in name {self.name_string}.")

    def get_standard_grid_name(self) -> str:
        return f"{self.algo}_{self.N}"

    def get_alg(self):
        return self.algo

    def get_N(self):
        return self.N

    def get_dim(self):
        return self.dim


"""
User script for generating pseudo-trajectories.
"""

import argparse

from molgri.constants import EXTENSION_TOPOLOGY, EXTENSION_TRAJECTORY
from molgri.molecules.writers import PtIOManager
from molgri.scripts.set_up_io import freshly_create_all_folders

import warnings
warnings.filterwarnings("ignore")

# TODO: define total_N and generate in all dimensions uniform grid?

parser = argparse.ArgumentParser()
requiredNamed = parser.add_argument_group('required named arguments')
requiredNamed.add_argument('-m1', type=str, nargs='?', required=True,
                           help='name of the .gro file containing the fixed molecule')
requiredNamed.add_argument('-m2', type=str, nargs='?', required=True,
                           help='name of the .gro file containing the mobile molecule')
requiredNamed.add_argument('-origingrid', metavar='og', type=str, nargs='?', required=True,
                           help='name of the rotation grid for rotations around origin in the form '
                                'algorithm_N (eg. ico_50) OR just a number (eg. 50)')
requiredNamed.add_argument('-bodygrid', metavar='bg', type=str, nargs='?', required=True,
                           help='name of the rotation grid for rotations around body in the form '
                                'algorithm_N (eg. ico_50) OR just a number (eg. 50) '
                                'OR None if you only want rotations about origin')
requiredNamed.add_argument('-transgrid', metavar='tg', type=str, nargs='?', required=True,
                           help='translation grid provided as a list of distances, as linspace(start, stop, num) '
                                'or range(start, stop, step) in nanometers')
parser.add_argument('--recalculate', action='store_true',
                    help='recalculate the grid even if a saved version already exists')
parser.add_argument('--only_origin', action='store_true',
                    help='An outdated function to suppress body rotations. We suggest using -b zero instead')
parser.add_argument('--as_dir', action='store_true',
                    help='Save the PT as a directory of frames')
parser.add_argument('--extension_trajectory', type=str, default=EXTENSION_TRAJECTORY,
                    help=f"File extension for generated (pseudo)-trajectories [default: {EXTENSION_TRAJECTORY}]")
parser.add_argument('--extension_structure', type=str, default=EXTENSION_TOPOLOGY,
                    help=f"File extension for generated topologies [default: {EXTENSION_TOPOLOGY}]")
parser.add_argument('--save_as', type=str, default=None, help='define the (base) name of output file')


def run_generate_pt():
    """

    Returns:

    """
    freshly_create_all_folders()
    my_args = parser.parse_args()
    if my_args.only_origin:
        my_args.bodygrid = "zero"
        print("Warning: the flag --only_origin is deprecated. We suggest setting -b zero instead.")
    if "ico" in my_args.bodygrid or "cube3D" in my_args.bodygrid:
        print(f"Warning! You are using -b {my_args.bodygrid} to create the grid of rotations around the COM. "
              f"We do not recommend the use of Icosahedron or 3D cube grids for this purpose; they are optimised "
              f"for generation of origin rotations (-o). We suggest using a 4D cube grid (cube4D) instead.")
    manager = PtIOManager(name_central_molecule=my_args.m1, name_rotating_molecule=my_args.m2,
                          output_name=my_args.save_as, b_grid_name=my_args.bodygrid, o_grid_name=my_args.origingrid,
                          t_grid_name=my_args.transgrid)
    manager.construct_pt_and_time(as_dir=my_args.as_dir, extension_trajectory=my_args.extension_trajectory,
                                  extension_structure=my_args.extension_structure, print_messages=True)


if __name__ == '__main__':
    run_generate_pt()




"""
Plots connected with (pseudo)trajectories and their energies.

Enables plotting COM of ParsedTrajectory and coloring them according to energies, convergence of energies with the
length of the (pseudo)trajectory.
"""

import numpy as np
import matplotlib.cm as cm
from matplotlib.colors import Normalize
import seaborn as sns

from molgri.constants import ENERGY2SHORT
from molgri.molecules.parsers import ParsedTrajectory
from molgri.plotting.abstract import RepresentationCollection, MultiRepresentationCollection
from molgri.plotting.fullgrid_plots import FullGridPlot


class TrajectoryPlot(RepresentationCollection):

    def __init__(self, parsed_trajectory: ParsedTrajectory, N_used: int = None):
        self.parsed_trajectory = parsed_trajectory
        if N_used is None:
            N_used = None
        self.N_used = N_used
        data_name = self.parsed_trajectory.get_name()
        super().__init__(data_name)

    def _default_atom_selection(self, atom_selection: str):
        if atom_selection is None and self.parsed_trajectory.is_pt:
            atom_selection = self.parsed_trajectory.get_atom_selection_r()
        return atom_selection

    def _make_scatter_plot(self, projection, data, **kwargs):
        sc = self.ax.scatter(*data, **kwargs)
        if projection == "3d":
            self._equalize_axes()
        elif projection == "hammer":
            self.ax.set_xticks([])
        return sc

    def get_possible_title(self):
        return f"N={self.N_used}"

    def make_COM_plot(self, ax=None, fig=None, save=True, atom_selection=None, projection="3d", animate_rot=False,
                      fg = None):
        self._create_fig_ax(ax=ax, fig=fig, projection=projection)
        # if a pseudotrajectory, default setting is to plot COM or r_molecule
        atom_selection = self._default_atom_selection(atom_selection)

        # filter out unique COMs
        coms, _ = self.parsed_trajectory.get_unique_com_till_N(N=self.N_used, atom_selection=atom_selection,
                                                               energy_type=None)
        if fg and projection == "3d":
            fgp = FullGridPlot(fg)
            fgp.make_full_voronoi_plot(ax=self.ax, fig=self.fig, animate_rot=False, plot_vertex_points=False,
                                       save=False, numbered=False)
            _, c = self.parsed_trajectory.assign_coms_2_grid_points(full_grid=fg, atom_selection=atom_selection,
                                                                 coms=coms)
            cmap="Spectral"
        else:
            c="black"
            cmap=None
        # plot data
        self._make_scatter_plot(projection, coms.T, c=c, cmap=cmap)

        if save:
            self._save_plot_type(f"com_{projection}")

        if animate_rot and projection == "3d":
            return self._animate_figure_view(self.fig, self.ax, f"com_rotated")

    def make_energy_COM_plot(self, ax=None, fig=None, save=True, atom_selection=None, projection="3d",
                             animate_rot=False, energy_type="Potential", vmin=None, vmax=None,
                             lowest_k = None, highest_j = None):
        self._create_fig_ax(ax=ax, fig=fig, projection=projection)
        atom_selection = self._default_atom_selection(atom_selection)

        coms, energies = self.parsed_trajectory.get_unique_com_till_N(N=self.N_used, energy_type=energy_type,
                                                                      atom_selection=atom_selection)
        coms, energies = self.parsed_trajectory.get_only_lowest_highest(coms, energies, lowest_k=lowest_k,
                                                                        highest_j=highest_j)
        # convert energies to colors
        cmap = cm.coolwarm
        if vmin is None:
            vmin = np.min(energies)
        if vmax is None:
            vmax = np.max(energies)
        norm = Normalize(vmin=vmin, vmax=vmax)

        # plot data
        sc = self._make_scatter_plot(projection, coms.T, cmap=cmap, c=energies, norm=norm)

        try:
            save_name = f"energies_{ENERGY2SHORT[energy_type]}"
        except KeyError:
            save_name = "energies"
        if save:
            self.fig.colorbar(sc, ax=self.ax)
            self._save_plot_type(f"{save_name}_{projection}")

        if animate_rot and projection == "3d":
            self._animate_figure_view(self.fig, self.ax, f"{save_name}_rotated")

    def make_molecule_plot(self):
        pass

    # kwargs are strictly necessary!
    def make_energy_violin_plot(self, ax=None, fig=None, save=True, energy_type: str = "Potential", **kwargs):
        self._create_fig_ax(ax=ax, fig=fig)
        coms, energies = self.parsed_trajectory.get_unique_com_till_N(N=self.N_used, energy_type=energy_type)
        sns.violinplot(energies, ax=self.ax, scale="count", inner="stick", cut=0)
        self.ax.set_xticklabels([])
        #self.ax.set_xlabel("N")
        self.ax.set_ylabel(f"{energy_type} [kJ/mol]")

        save_name = f"violin_energies_{ENERGY2SHORT[energy_type]}"
        if save:
            self._save_plot_type(save_name)

    def create_all_plots(self, and_animations=False):
        self.make_COM_plot(projection="3d", animate_rot=and_animations)
        self.make_COM_plot(projection="hammer")
        for energy_type in self.parsed_trajectory.energies.labels:
            self.make_energy_COM_plot(projection="3d", animate_rot=and_animations, energy_type=energy_type)
            self.make_energy_COM_plot(projection="hammer", energy_type=energy_type)
            self.make_energy_violin_plot(energy_type=energy_type)


class ConvergenceMultiCollection(MultiRepresentationCollection):

    def __init__(self, parsed_trajectory: ParsedTrajectory, N_set: tuple = None):
        data_name = f"convergence_{parsed_trajectory.get_name()}"

        if N_set is None:
            N_max = parsed_trajectory.get_num_unique_com()
            N_set = np.linspace(N_max//10, N_max, num=5, dtype=int)
        self.N_set = N_set

        # generate TrajectoryPlots with limited number of points
        list_plots = [TrajectoryPlot(parsed_trajectory, N_used=N) for N in self.N_set]
        super().__init__(data_name, list_plots, n_columns=len(self.N_set), n_rows=1)

    def make_all_COM_3d_plots(self, animate_rot=False, save=True):

        self._make_plot_for_all("make_COM_plot", projection="3d", remove_midlabels=False,
                                creation_kwargs={"sharex": False, "sharey": False},
                                plotting_kwargs={"projection": "3d"})

        titles = [subplot.get_possible_title() for subplot in self.list_plots]
        self.add_titles(list_titles=titles, pad=-14)
        self.unify_axis_limits()

        if animate_rot:
            self.animate_figure_view(f"COM_3d")
        if save:
            self._save_multiplot(f"COM_3d")

    def make_all_COM_hammer_plots(self, save=True):
        self._make_plot_for_all("make_COM_plot", projection="hammer", remove_midlabels=False,
                                creation_kwargs={"sharex": False, "sharey": False},
                                plotting_kwargs={"projection": "hammer"})

        titles = [subplot.get_possible_title() for subplot in self.list_plots]
        self.add_titles(list_titles=titles)

        if save:
            self._save_multiplot(f"COM_hammer")

    def make_all_energy_plots(self, dim, animate_rot=False, save=True, energy_type: str = "Potential"):
        _, all_energies = self.list_plots[-1].parsed_trajectory.get_unique_com_till_N(N=np.max(self.N_set),
                                                                                   energy_type=energy_type)
        vmax = np.max(all_energies)
        vmin = np.min(all_energies)
        if dim == 3:
            projection = "3d"
            method = "make_energy_COM_plot"
            sharey = False
            pad = -14
        elif dim == 2:
            projection = "hammer"
            method = "make_energy_COM_plot"
            sharey = False
            pad = 0
        elif dim == 1:
            projection = None
            method = "make_energy_violin_plot"
            sharey = True
            pad = 0
        else:
            raise ValueError("Only energy plots with 1, 2 or 3 dimensions possible!")

        self._make_plot_for_all(method, projection=projection, remove_midlabels=False,
                                creation_kwargs={"sharex": False, "sharey": sharey},
                                plotting_kwargs={"projection": projection, "energy_type": energy_type,
                                                 "vmin": vmin, "vmax": vmax})

        titles = [subplot.get_possible_title() for subplot in self.list_plots]
        self.add_titles(list_titles=titles, pad=pad)

        if dim == 3:
            self.unify_axis_limits()
            self.add_colorbar()
        if dim == 2:
            self.add_colorbar()
        if dim == 1:
            self.unify_axis_limits(x_ax=False, y_ax=True)

        if animate_rot and dim == 3:
            self.animate_figure_view(f"energy_{dim}d", dpi=100)
        if save:
            self._save_multiplot(f"energy_{dim}d")

    def create_all_plots(self, and_animations=False):
        self.make_all_COM_3d_plots(animate_rot=and_animations)
        self.make_all_COM_hammer_plots()
        energy_types = self.list_plots[0].parsed_trajectory.energies.labels
        for energy_type in energy_types:
            for dim in (1, 2, 3):
                self.make_all_energy_plots(dim, energy_type=energy_type)



import numpy as np
from time import time
import pandas as pd
import subprocess
import seaborn as sns
import matplotlib.pyplot as plt

from molgri.space.translations import TranslationParser
from molgri.paths import PATH_OUTPUT_TIMING, PATH_OUTPUT_PLOTS
from molgri.constants import DIM_PORTRAIT, DEFAULT_DPI, DIM_LANDSCAPE

import timeit
import re
import subprocess as sp

def execute_scripts(script_setup):
    subprocess.call(script_setup)




def time_pt_creation(mol1, mol2, distances="[5, 10, 15]", N_rot_max=100, num_points=30):
    """
    Create PTs with the two provided molecules
    Args:
        mol1:
        mol2:
        distances:
        N_rot_max:

    Returns:

    """
    radii = TranslationParser(distances)
    num_samples = 5
    N_t = len(radii.get_trans_grid())
    N_rot_selections = np.linspace(1, N_rot_max, num=num_points, dtype=int)
    data = np.zeros((2, num_samples*len(N_rot_selections)))
    current_index = 0
    for N_rot in N_rot_selections:
        # perform molgri-pt
        commands = ["python", "-m", "molgri.scripts.generate_pt", "-m1", f"{mol1}", "-m2", f"{mol2}",
                        "-t", f"{distances}", "-o", f"{N_rot}", "-b", f"{N_rot}", "--recalculate"]
        t = timeit.Timer(f"execute_scripts({commands})", setup="from __main__ import execute_scripts")

        for _ in range(num_samples):
            # save data
            data[0][current_index] = N_rot * N_rot * N_t
            data[1][current_index] = t.timeit(1) / 60
            current_index += 1

    df = pd.DataFrame(data.T, columns=["Number of PT frames", "Generation time [min]"])
    print(df)
    df.to_csv(f"{PATH_OUTPUT_TIMING}{mol1}_{mol2}_time_{num_points}.csv")


def plot_timing(f_names):
    sns.set_context("talk")
    colors = ["red", "blue"]
    fig, ax = plt.subplots(1, 1, figsize=DIM_LANDSCAPE)
    for f_name, color in zip(f_names, colors):
        df = pd.read_csv(f_name)
        sns.lineplot(df, x="Number of PT frames", y="Generation time [min]", color=color, ax=ax)
    ax.set_xticks([100*100*3, 300*300*3, 500*500*3])
    xlabels = [f'{100*100*3}\n(100x100x3)', f'{300*300*3}\n(300x300x3)', f'{500*500*3}\n(500x500x3)']
    #xlabels_new = [re.sub("(.{8})", "\\1\n", label, 0, re.DOTALL) for label in xlabels]

    ax.set_xticklabels(xlabels)
    plt.tight_layout()
    plt.savefig(f"{PATH_OUTPUT_PLOTS}timing_pres.pdf", dpi=DEFAULT_DPI)
    #plt.show()

if __name__ == "__main__":
    num_points = 30
    #time_pt_creation("example_protein", "example_protein", N_rot_max=500, num_points=num_points)
    #time_pt_creation("example_pdb", "example_pdb", N_rot_max=500, num_points=num_points)
    #plot_timing(f_name=)
    #f"{PATH_OUTPUT_TIMING}example_pdb_example_pdb_time_{num_points}.csv",
    plot_timing([
                 f"{PATH_OUTPUT_TIMING}example_protein_example_protein_time_{num_points}.csv"])



"""
Conversions rotations <-> grid points.
"""

from typing import Tuple

import numpy as np
from numpy.typing import NDArray
from scipy.spatial.transform import Rotation

from molgri.space.utils import normalise_vectors


def rotation2grid(rotations: Rotation) -> Tuple[NDArray, ...]:
    """
    Convert a series of N rotational objects (represented as a scipy object Rotation) to three grids by applying
    the rotations to a unit vectors in x, y, z directions.  The grids can be saved and converted to a rotational
    object later if needed or used in grid form to get positional grids in spherical coordinates.

    Args:
        rotations: a series of N rotational objects (represented as a scipy object Rotation)

    Returns:
        a tuple of three numpy arrays, each of shape (N, 3)
    """

    basis = np.eye(3)

    result = []
    for basis_vector in basis:
        rotated_bv = rotation2grid4vector(rotations, basis_vector)
        # special case when only one rotation is applied - need to expand dimensions
        if rotated_bv.shape == (3,):
            rotated_bv = rotated_bv[np.newaxis, :]
        assert rotated_bv.shape[1] == 3
        result.append(rotated_bv)
    return tuple(result)


def rotation2grid4vector(rotations: Rotation, vector: NDArray = None) -> NDArray:
    """
    Convert a set of rotations to a set of grid points by applying the rotation(s) to a specific vector.
    """
    if vector is None:
        vector = np.array([0, 0, 1])
    return rotations.apply(vector)


def quaternion2grid(quaternions: NDArray) -> Tuple[NDArray, ...]:
    """
    See rotation2grid function. This is only a helper function that parsers quaternions as inputs.
    """
    assert quaternions.shape[1] == 4, "Quaternions must have 4 dimensions"
    rotations = Rotation.from_quat(quaternions)
    return rotation2grid(rotations)


def grid2rotation(grid_x: NDArray, grid_y: NDArray, grid_z: NDArray) -> Rotation:
    """
    Re-create a rotational object by using the (saved) grid_x, grid_y and grid_z projections. We are looking for an
    array of rotational matrices R that achieve
        R[i] (1, 0, 0)^T = grid_x[i]
        R[i] (0, 1, 0)^T = grid_y[i]
        R[i] (0, 0, 1)^T = grid_z[i]
    for each i in range(len(grids)). It is easy to show that

            grid_x[i][0]  grid_y[i][0]  grid_z[i][0]
    R[i] =  grid_x[i][1]  grid_y[i][1]  grid_z[i][1]
            grid_x[i][2]  grid_y[i][2]  grid_z[i][2]

    Args:
        grid_x: an array (N, 3) where each row is a coordinate on a 3D sphere created by projecting rotation on 1, 0, 0
        grid_y: an array (N, 3) where each row is a coordinate on a 3D sphere created by projecting rotation on 0, 1, 0
        grid_z: an array (N, 3) where each row is a coordinate on a 3D sphere created by projecting rotation on 0, 0, 1

    Returns:
        a list of length N where each element is a rotational object
    """
    assert np.allclose(np.linalg.norm(grid_x, axis=1), 1), "Points on a x grid must be unit vectors!"
    assert np.allclose(np.linalg.norm(grid_y, axis=1), 1), "Points on a y grid must be unit vectors!"
    assert np.allclose(np.linalg.norm(grid_z, axis=1), 1), "Points on a z grid must be unit vectors!"
    assert grid_x.shape[1] == 3, f"grid_x must be of shape (N, 3), not {grid_x.shape}"
    assert grid_y.shape[1] == 3, f"grid_y must be of shape (N, 3), not {grid_y.shape}"
    assert grid_z.shape[1] == 3, f"grid_z must be of shape (N, 3), not {grid_z.shape}"

    # re-create rotational matrices from the three directional grids
    rot_matrices = np.concatenate((grid_x, grid_y, grid_z), axis=1)
    rot_matrices = rot_matrices.reshape((-1, 3, 3))
    rot_matrices = rot_matrices.swapaxes(1, 2)
    assert len(rot_matrices) == len(grid_x) == len(grid_y) == len(grid_z)
    # rot_matrices is an array in which each "row" is a 3x3 rotational matrix
    # create a rotational object set from a stack of rotational matrices
    rotations = Rotation.from_matrix(rot_matrices)
    return rotations


def grid2quaternion(grid_x: NDArray, grid_y: NDArray, grid_z: NDArray) -> NDArray:
    """
    See grid2rotation; this function only reformats the output as a (N, 4) array of quaternions.
    """
    rot_objects = grid2rotation(grid_x, grid_y, grid_z)
    quaternions = rot_objects.as_quat()
    assert quaternions.shape[1] == 4, "quaternions must have 4 dimensions"
    return quaternions


# ########################## HELPER FUNCTIONS ################################

def skew(x: NDArray) -> NDArray:
    """
    Take a vector or an array of vectors and return its skew matrix/matrices.

    Args:
        x: a vector (3,) or (N, 3)

    Returns:
        skew matrix, see structure below
    """

    def skew_2D(m):
        """Implement the definition of the skew matrix."""
        return np.array([[0, -m[2], m[1]],
                         [m[2], 0, -m[0]],
                         [-m[1], m[0], 0]])

    assert x.shape == (3,) or (len(x.shape) == 2 and x.shape[1] == 3)

    if x.shape == (3,):
        return skew_2D(x)
    else:
        skew_matrix = np.zeros((len(x), 3, 3))
        for i, row in enumerate(x):
            skew_matrix[i] = skew_2D(row)
        return skew_matrix


def two_vectors2rot(x: NDArray, y: NDArray) -> NDArray:
    """
    Take vectors x and y (or arrays of vectors with the same number of elements and return the
    rotational matrix that transforms x into y.

    Args:
        x: an array of shape (3,), first vector, or an array of vectors of size (N, 3)
        y: an array of shape (3,),  second vector, or an array of vectors of size (N, 3)

    Returns:
        a 3x3 rotational matrix
    """
    # checking the inputs
    assert y.shape == x.shape, "Arrays must be of same size."
    if len(x.shape) == 1:
        assert x.shape == (3,)
    elif len(x.shape) == 2:
        assert x.shape[1] == 3
    else:
        raise AttributeError("Arrays must be 1D or 2D")

    # converting 1D vectors in (1, 3) shape
    if y.shape == x.shape == (3,):
        x = x[np.newaxis, :]
        y = y[np.newaxis, :]
        assert y.shape == x.shape == (1, 3)
    N = len(x)

    # if not normalised yet, normalise
    x = normalise_vectors(x, axis=1)
    y = normalise_vectors(y, axis=1)
    assert np.allclose(np.linalg.norm(x, axis=1), 1) and np.allclose(np.linalg.norm(y, axis=1), 1)

    v = np.cross(x, y)
    s = np.linalg.norm(v, axis=1)
    c = np.matmul(x, y.T)
    c = c.diagonal()
    np.seterr(all="ignore")
    factor = np.divide((1 - c), s ** 2)
    my_matrix = N_eye_matrices(N, d=3) + skew(v) + np.matmul(skew(v), skew(v)) * factor[:, np.newaxis, np.newaxis]
    np.seterr(all="warn")
    sinus_zero = s[:, np.newaxis, np.newaxis] == 0
    cosinus_one = c[:, np.newaxis, np.newaxis] == 1
    my_matrix = np.where(np.logical_and(sinus_zero, cosinus_one), N_eye_matrices(N, d=3), my_matrix)
    my_matrix = np.where(np.logical_and(sinus_zero, ~cosinus_one),
                         -N_eye_matrices(N, d=3), my_matrix)
    if len(x) == len(y) == 1:
        my_matrix = my_matrix[0]
        assert my_matrix.shape == (3, 3)
    else:
        assert my_matrix.shape == (len(x), 3, 3)
    return my_matrix


def N_eye_matrices(N: int, d: int = 3):
    """
    Returns a (N, d, d) array in which each 'row' is an identity matrix.

    Args:
        N: number of rows
        d: dimension of the identity matrix

    Returns:
        a "hyper"-matrix of identity matrices
    """
    shape = (N, d, d)
    identity_d = np.zeros(shape)
    idx = np.arange(shape[1])
    identity_d[:, idx, idx] = 1
    return identity_d



"""
Plotting SphereGridNDim and Polytope-based objects.

Visualising 3D with 3D and hammer plots and 4D with translation animation and cell plots. Plotting spherical Voronoi
cells and their areas. A lot of convergence and uniformity testing.
"""

import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.pyplot import Figure, Axes
from mpl_toolkits.mplot3d.axes3d import Axes3D
from matplotlib.animation import FuncAnimation
from numpy.typing import NDArray
from seaborn import color_palette
import networkx as nx

from molgri.constants import DEFAULT_ALPHAS_3D, TEXT_ALPHAS_3D, DEFAULT_ALPHAS_4D, TEXT_ALPHAS_4D, COLORS, \
    GRID_ALGORITHMS, NAME2SHORT_NAME, DEFAULT_DPI_MULTI, DIM_LANDSCAPE, DIM_SQUARE, DEFAULT_NS
from molgri.plotting.abstract import RepresentationCollection, PanelRepresentationCollection, plot_voronoi_cells
from molgri.space.analysis import vector_within_alpha
from molgri.space.polytopes import Polytope, second_neighbours, third_neighbours, PolyhedronFromG
from molgri.space.rotobj import SphereGridNDim, SphereGridFactory, ConvergenceSphereGridFactory


class SphereGridPlot(RepresentationCollection):

    def __init__(self, sphere_grid: SphereGridNDim, **kwargs):
        data_name = sphere_grid.get_name(with_dim=True)
        super().__init__(data_name, default_axes_limits=(-1, 1, -1, 1, -1, 1), **kwargs)
        self.sphere_grid = sphere_grid
        if self.sphere_grid.dimensions == 3:
            self.alphas = DEFAULT_ALPHAS_3D
            self.alphas_text = TEXT_ALPHAS_3D
        else:
            self.alphas = DEFAULT_ALPHAS_4D
            self.alphas_text = TEXT_ALPHAS_4D

    def get_possible_title(self):
        return NAME2SHORT_NAME[self.sphere_grid.algorithm_name]

    def make_grid_plot(self, fig: Figure = None, ax: Axes3D = None, save: bool = True):
        """Plot the 3D grid plot, for 4D the 4th dimension plotted as color. It always has limits (-1, 1) and equalized
        figure size"""

        self._create_fig_ax(fig=fig, ax=ax, projection="3d")

        points = self.sphere_grid.get_grid_as_array()
        if points.shape[1] == 3:
            self.ax.scatter(*points.T, color="black", s=30)
        else:
            self.ax.scatter(*points[:, :3].T, c=points[:, 3].T, s=30)  # cmap=plt.hot()

        self.ax.view_init(elev=10, azim=30)
        self._set_axis_limits()
        self._equalize_axes()

        if save:
            self._save_plot_type("grid")

    def make_grid_colored_with_alpha(self, ax=None, fig=None, central_vector: NDArray = None, save=True):
        if self.sphere_grid.dimensions != 3:
            print(f"make_grid_colored_with_alpha currently implemented only for 3D systems.")
            return None
        if central_vector is None:
            central_vector = np.zeros((self.sphere_grid.dimensions,))
            central_vector[-1] = 1

        self._create_fig_ax(fig=fig, ax=ax, projection="3d")

        points = self.sphere_grid.get_grid_as_array()

        # plot vector
        self.ax.scatter(*central_vector, marker="x", c="k", s=30)
        # determine color palette
        cp = sns.color_palette("Spectral", n_colors=len(self.alphas))
        # sort points which point in which alpha area
        already_plotted = []
        for i, alpha in enumerate(self.alphas):
            possible_points = np.array([vec for vec in points if tuple(vec) not in already_plotted])
            within_alpha = vector_within_alpha(central_vector, possible_points, alpha)
            selected_points = [tuple(vec) for i, vec in enumerate(possible_points) if within_alpha[i]]
            array_sel_points = np.array(selected_points)
            if np.any(array_sel_points):
                self.ax.scatter(*array_sel_points.T, color=cp[i], s=30)
            already_plotted.extend(selected_points)

        self.ax.view_init(elev=10, azim=30)
        self._set_axis_limits()
        self._equalize_axes()

        if save:
            self._save_plot_type("colorful_grid")
        return fig, ax

    def make_uniformity_plot(self, ax: Axes = None, fig: Figure = None, save=True):
        """
        Creates violin plots that are a measure of grid uniformity. A good grid will display minimal variation
        along a range of angles alpha.
        """

        self._create_fig_ax(fig=fig, ax=ax)

        df = self.sphere_grid.get_uniformity_df(alphas=self.alphas)
        sns.violinplot(x=df["alphas"], y=df["coverages"], ax=self.ax, palette=COLORS, linewidth=1, scale="count", cut=0)
        self.ax.set_xticklabels(self.alphas_text)

        if save:
            self._save_plot_type("uniformity")
        return self.fig, self.ax

    def make_convergence_plot(self, ax: Axes = None, fig: Figure = None, save=True):
        """
        Creates convergence plots that show how coverages approach optimal values.
        """
        self._create_fig_ax(fig=fig, ax=ax)
        df = self.sphere_grid.get_convergence_df(alphas=self.alphas)
        sns.lineplot(x=df["N"], y=df["coverages"], ax=self.ax, hue=df["alphas"],
                     palette=color_palette("hls", len(self.alphas_text)), linewidth=1)
        sns.lineplot(x=df["N"], y=df["ideal coverage"], style=df["alphas"], ax=self.ax, color="black")

        self.ax.get_legend().remove()

        if save:
            self.ax.set_xscale("log")
            self.ax.set_yscale("log")
            self._save_plot_type("convergence")

    def make_spherical_voronoi_plot(self, ax=None, fig=None, save=True, animate_rot=False):

        if self.sphere_grid.dimensions != 3:
            print("make_spherical_voronoi_plot only implemented for 3D grids")
            return

        self._create_fig_ax(fig=fig, ax=ax, projection="3d")

        try:
            sv = self.sphere_grid.get_spherical_voronoi_cells()
            plot_voronoi_cells(sv, self.ax)
        except AttributeError:
            pass

        self.ax.view_init(elev=10, azim=30)
        self._set_axis_limits()
        self._equalize_axes()

        if save:
            self._save_plot_type("sph_voronoi")
        if animate_rot:
            return self._animate_figure_view(self.fig, self.ax, f"sph_voronoi_rotated")

    def make_rot_animation(self, **kwargs):
        self.make_grid_plot(save=False)
        return self._animate_figure_view(self.fig, self.ax, **kwargs)

    def make_ordering_animation(self):
        self.make_grid_plot(save=True)
        sc = self.ax.collections[0]

        def update(i):
            current_colors = np.concatenate([facecolors_before[:i], all_white[i:]])
            sc.set_facecolors(current_colors)
            sc.set_edgecolors(current_colors)
            return sc,

        facecolors_before = sc.get_facecolors()
        shape_colors = facecolors_before.shape
        all_white = np.zeros(shape_colors)

        self.ax.view_init(elev=10, azim=30)
        ani = FuncAnimation(self.fig, func=update, frames=len(facecolors_before), interval=100, repeat=False)
        fps_factor = np.min([len(facecolors_before), 20])
        self._save_animation_type(ani, "order", fps=len(facecolors_before) // fps_factor)
        return ani

    def make_trans_animation(self, fig: plt.Figure = None, ax=None, save=True, **kwargs):
        points = self.sphere_grid.get_grid_as_array()
        # create the axis with the right num of dimensions

        dimension = points.shape[1] - 1

        if dimension == 3:
            projection = "3d"
        else:
            projection = None

        self._create_fig_ax(fig=fig, ax=ax, projection=projection)
        # sort by the value of the specific dimension you are looking at
        ind = np.argsort(points[:, -1])
        points_3D = points[ind]
        # map the 4th dimension into values 0-1
        alphas = points_3D[:, -1].T
        alphas = (alphas - np.min(alphas)) / np.ptp(alphas)

        # plot the lower-dimensional scatterplot
        all_points = []
        for line in points_3D:
            all_points.append(self.ax.scatter(*line[:-1], color="black", alpha=1))

        self._set_axis_limits((-1, 1) * dimension)
        self._equalize_axes()

        if len(points) < 20:
            step = 1
        elif len(points) < 100:
            step = 5
        else:
            step = 20

        def animate(frame):
            # plot current point
            current_time = alphas[frame * step]
            for i, p in enumerate(all_points):
                new_alpha = np.max([0, 1 - np.abs(alphas[i] - current_time) * 10])
                p.set_alpha(new_alpha)
            return self.ax,

        anim = FuncAnimation(self.fig, animate, frames=len(points) // step,
                             interval=100)  # , frames=180, interval=50
        if save:
            self._save_animation_type(anim, "trans", fps=len(points) // step // 2, **kwargs)
        return anim

    def create_all_plots(self, and_animations=False):
        self.make_grid_plot()
        self.make_grid_colored_with_alpha()
        self.make_uniformity_plot()
        self.make_convergence_plot()
        self.make_spherical_voronoi_plot(animate_rot=and_animations)
        if and_animations:
            self.make_rot_animation()
            self.make_ordering_animation()
            self.make_trans_animation()


class PolytopePlot(RepresentationCollection):

    def __init__(self, polytope: Polytope, **kwargs):
        self.polytope = polytope
        split_name = str(polytope).split()
        data_name = f"{split_name[0]}_{split_name[-1]}"
        default_complexity_level = kwargs.pop("default_complexity_level", "half_empty")
        super().__init__(data_name, default_complexity_level=default_complexity_level, **kwargs)

    def make_graph(self, ax=None, fig=None, with_labels=True, save=True):
        """
        Plot the networkx graph of self.G.
        """
        self._create_fig_ax(fig=fig, ax=ax)

        node_labels = {i: tuple(np.round(i, 3)) for i in self.polytope.G.nodes}
        nx.draw_networkx(self.polytope.G, pos=nx.spring_layout(self.polytope.G, weight="p_dist"),
                         with_labels=with_labels, labels=node_labels)
        if save:
            self._save_plot_type("graph")

    def make_neighbours_plot(self, ax: Axes3D = None, fig: Figure = None, save: bool = True, node_i: int = 0):
        """
        Want to see which points count as neighbours, second- or third neighbours of a specific node? Use this plotting
        method.
        """

        if self.polytope.d != 3:
            print(f"Plotting neighbours not available for d={self.polytope.d}")
            return

        self._create_fig_ax(fig=fig, ax=ax, projection="3d")

        all_nodes = self.polytope.get_node_coordinates()
        node = tuple(all_nodes[node_i])
        neig = self.polytope.G.neighbors(node)
        sec_neig = list(second_neighbours(self.polytope.G, node))
        third_neig = list(third_neighbours(self.polytope.G, node))
        for sel_node in all_nodes:

            self.ax.scatter(*sel_node, color="black", s=30, alpha=0.5)
            if np.allclose(sel_node, node):
                self.ax.scatter(*sel_node, color="red", s=40, alpha=0.5)
            if tuple(sel_node) in neig:
                self.ax.scatter(*sel_node, color="blue", s=38, alpha=0.5)
            if tuple(sel_node) in sec_neig:
                self.ax.scatter(*sel_node, color="green", s=35, alpha=0.5)
            if tuple(sel_node) in third_neig:
                self.ax.scatter(*sel_node, color="orange", s=32, alpha=0.5)
        if save:
            self._save_plot_type(f"neighbours_{node_i}")

    def make_node_plot(self, ax: Axes3D = None, fig: Figure = None, select_faces: set = None, projection: bool = False,
                       plot_edges: bool = False, plot_nodes: bool = True,
                       color_by: str = "level", enumerate_nodes=False, save=True):
        """
        Plot the points of the polytope + possible division points. Colored by level at which the point was added.
        Or: colored by index to see how sorting works. Possible to select only one or a few faces on which points
        are to be plotted for clarity.

        Args:
            fig: figure
            ax: axis
            save: whether to save fig
            select_faces: a set of face numbers that can range from 0 to number of faces of the polyhedron, e.g. {0, 5}.
                          If None, all faces are shown.
            projection: True if you want to plot the projected points, not the ones on surfaces of polytope
            plot_edges: select True if you want to see connections between nodes
            color_by: "level" or "index"
        """
        if self.polytope.d > 3:
            print(f"Plotting nodes not available for d={self.polytope.d}")
            return

        self._create_fig_ax(fig=fig, ax=ax, projection="3d")

        level_color = ["black", "red", "blue", "green"]
        index_palette = color_palette("coolwarm", n_colors=self.polytope.G.number_of_nodes())

        if plot_nodes:
            for i, point in enumerate(self.polytope.get_N_ordered_points(projections=False)):
                # select only points that belong to at least one of the chosen select_faces (or plot all if None selection)
                node = self.polytope.G.nodes[tuple(point)]
                point_faces = set(node["face"])
                point_level = node["level"]
                point_projection = node["projection"]
                if select_faces is None or len(point_faces.intersection(select_faces)) > 0:
                    # color selected based on the level of the node or index of the sorted nodes
                    if color_by == "level":
                        color = level_color[point_level]
                    elif color_by == "index":
                        color = index_palette[i]
                    else:
                        raise ValueError(f"The argument color_by={color_by} not possible (try 'index', 'level')")

                    if projection:
                        self.ax.scatter(*point_projection, color=color, s=30)
                    else:
                        self.ax.scatter(*point, color=color, s=30)
                        if enumerate_nodes:
                            self.ax.text(*point, s=f"{i}")
        self._set_axis_limits((-0.6, 0.6, -0.6, 0.6, -0.6, 0.6))
        self._equalize_axes()
        if plot_edges:
            self._plot_edges(self.ax, select_faces=select_faces)
        if save:
            self._save_plot_type(f"points_{color_by}")

    def _plot_edges(self, ax, select_faces=None, label=None, **kwargs):
        """
        Plot the edges between the points. Can select to display only some faces for clarity.

        Args:
            ax: axis
            select_faces: a set of face numbers from 0 to (incl) 19, e.g. {0, 5}. If None, all faces are shown.
            label: select the name of edge parameter if you want to display it
            **kwargs: other plotting arguments
        """
        for edge in self.polytope.G.edges(data=True):
            faces_edge_1 = set(self.polytope.G.nodes[edge[0]]["face"])
            faces_edge_2 = set(self.polytope.G.nodes[edge[1]]["face"])
            # both the start and the end point of the edge must belong to one of the selected faces
            n1_on_face = select_faces is None or len(faces_edge_1.intersection(select_faces)) > 0
            n2_on_face = select_faces is None or len(faces_edge_2.intersection(select_faces)) > 0
            if n1_on_face and n2_on_face:
                # usually you only want to plot edges used in division
                ax.plot(*np.array(edge[:2]).T, color="black",  **kwargs)
                if label:
                    midpoint = np.average(np.array(edge[:2]), axis=0)
                    s = edge[2][f"{label}"]
                    ax.text(*midpoint, s=f"{s:.3f}")

    def make_cell_plot(self, ax: Axes3D = None, fig: Figure = None, cell_index: int = 0, draw_edges: bool = True,
                       save: bool = True, animate_rot: bool = False, plot_nodes: bool = True, projection = False):
        """
        Since you cannot visualise a 4D object directly, here's an option to visualise the 3D sub-cells of a 4D object.

        Args:
            fig: figure
            ax: axis
            save: whether to save fig
            cell_index: index of the sub-cell to plot (in cube4D that can be 0-7)
            draw_edges: use True if you want to also draw edges, False if only points
        """
        if self.polytope.d != 4:
            print(f"Plotting cells not available for d={self.polytope.d}")
            return

        self._create_fig_ax(ax=ax, fig=fig, dim=3, projection="3d")
        # find the points that belong to the chosen cell_index
        nodes = (
            node
            for node, data
            in self.polytope.G.nodes(data=True)
            if cell_index in data.get('face')
        )
        subgraph = self.polytope.G.subgraph(nodes)
        # find the component corresponding to the constant 4th dimension
        arr_nodes = np.array(subgraph.nodes)
        dim_to_keep = list(np.where(~np.all(arr_nodes == arr_nodes[0, :], axis=0))[0])
        new_nodes = {old: (old[dim_to_keep[0]], old[dim_to_keep[1]], old[dim_to_keep[2]]) for old in subgraph.nodes}
        subgraph_3D = nx.relabel_nodes(subgraph, new_nodes)
        # create a 3D polyhedron and use its plotting functions
        sub_polyhedron = PolyhedronFromG(subgraph_3D)
        poly_plotter = PolytopePlot(sub_polyhedron)
        poly_plotter.make_node_plot(ax=self.ax, plot_edges=draw_edges, plot_nodes=plot_nodes, projection=projection)
        self._set_axis_limits((-0.6, 0.6, -0.6, 0.6, -0.6, 0.6))
        self._equalize_axes()
        if save:
            self._save_plot_type("cell")
        if animate_rot:
            return self._animate_figure_view(self.fig, self.ax, f"cell_rotated")

    def create_all_plots(self):
        self.make_graph(with_labels=True)
        self.make_neighbours_plot()
        self.make_node_plot(projection=True, color_by="level")
        self.make_node_plot(plot_edges=True, color_by="index")
        self.make_node_plot(select_faces={0, 1, 2})
        self.make_cell_plot()


class PanelSphereGridPlots(PanelRepresentationCollection):

    def __init__(self, N_points: int, grid_dim: int, default_context: str = None, default_complexity_level: str = None,
                 default_color_style: str = None, use_saved=False, **kwargs):
        list_plots = []
        for alg in GRID_ALGORITHMS[:-1]:
            sphere_grid = SphereGridFactory.create(alg_name=alg, N=N_points, dimensions=grid_dim, print_messages=False,
                                                   time_generation=False, filter_non_unique=True, use_saved=use_saved)
            sphere_plot = SphereGridPlot(sphere_grid, default_context=default_context,
                                         default_color_style=default_color_style,
                                         default_complexity_level=default_complexity_level)
            list_plots.append(sphere_plot)
        data_name = f"all_{N_points}_{grid_dim}d"
        super().__init__(data_name, list_plots, **kwargs)

    def make_all_grid_plots(self, animate_rot=False, animate_trans=False):
        self._make_plot_for_all("make_grid_plot", projection="3d")
        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots],
                        pad=-14)
        if animate_rot:
            self.animate_figure_view("grid", dpi=100)
        if animate_trans:
            self._make_plot_for_all("make_trans_animation", projection="3d")
        self._save_multiplot("grid", dpi=100)

    def make_all_trans_animations(self, fig: plt.Figure = None, ax=None, save=True, **kwargs):

        all_points = []
        for plot in self.list_plots:
            all_points.append(plot.sphere_grid.get_grid_as_array())


        dimension = all_points[0].shape[1] - 1

        if dimension == 3:
            projection = "3d"
        else:
            projection = None

        figsize = self.n_columns * DIM_SQUARE[0], self.n_rows * DIM_SQUARE[1]
        self.fig, self.all_ax = plt.subplots(self.n_rows, self.n_columns, subplot_kw={'projection': projection},
                                             figsize=figsize)

        # sort by the value of the specific dimension you are looking at
        all_points_3D = []
        for i, points in enumerate(all_points):
            ind = np.argsort(points[:, -1])
            all_points_3D.append(points[ind])
        # map the 4th dimension into values 0-1
        all_alphas = []
        for points_3D in all_points_3D:
            alphas = points_3D[:, -1].T
            alphas = (alphas - np.min(alphas)) / np.ptp(alphas)
            all_alphas.append(alphas)

        # plot the lower-dimensional scatterplot
        plotted_points = []
        for ax, points3D in zip(np.ravel(self.all_ax), all_points_3D):
            ax.set_box_aspect(aspect=[1, 1, 1])
            sub_plotted_points = []
            for line in points3D:
                sub_plotted_points.append(ax.scatter(*line[:-1], color="black", alpha=1))
            plotted_points.append(sub_plotted_points)

        self.all_ax[0, 0].set_xlim(-1, 1)
        self.all_ax[0, 0].set_ylim(-1, 1)
        if dimension == 3:
            self.all_ax[0, 0].set_zlim(-1, 1)
        self.unify_axis_limits()
        for ax in np.ravel(self.all_ax):
            ax.set_xticks([])
            ax.set_yticks([])
            if dimension == 3:
                ax.set_zticks([])

        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots],
                        pad=-14)

        if len(all_points_3D[0]) < 20:
            step = 1
        elif len(all_points_3D[0]) < 100:
            step = 5
        else:
            step = 20

        def animate(frame):
            # plot current point
            current_time = alphas[frame * step]
            for ax_alphas, ax_points, ax in zip(all_alphas, plotted_points, np.ravel(self.all_ax)):
                for i, p in enumerate(ax_points):
                    new_alpha = np.max([0, 1 - np.abs(ax_alphas[i] - current_time) * 10])
                    p.set_alpha(new_alpha)
            return self.all_ax,

        anim = FuncAnimation(self.fig, animate, frames=len(all_points_3D[0]) // step,
                             interval=100)  # , frames=180, interval=50
        if save:
            self.save_multianimation(anim, "trans", fps=len(all_points_3D[0]) // step // 2, dpi=200)
        return anim

    def make_all_convergence_plots(self):
        self._make_plot_for_all("make_convergence_plot")
        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots])
        self.set_log_scale()
        self.unify_axis_limits()
        self._save_multiplot("convergence")

    def make_all_uniformity_plots(self):
        self._make_plot_for_all("make_uniformity_plot")
        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots])
        self.unify_axis_limits()
        self._save_multiplot("uniformity")

    def create_all_plots(self, and_animations=False):
        self.make_all_grid_plots(animate_rot=and_animations)
        self.make_all_convergence_plots()
        self.make_all_uniformity_plots()


class ConvergenceSphereGridPlot(RepresentationCollection):

    def __init__(self, convergence_sph_grid: ConvergenceSphereGridFactory):
        self.convergence_sph_grid = convergence_sph_grid
        super().__init__(self.convergence_sph_grid.get_name())

    def get_possible_title(self):
        full_name = self.convergence_sph_grid.get_name()
        split_name = full_name.split("_")
        return NAME2SHORT_NAME[split_name[1]]

    def make_voronoi_area_conv_plot(self, ax=None, fig=None, save=True):
        if self.convergence_sph_grid.dimensions != 3:
            print(f"make_voronoi_area_conv_plot available only for 3D systems")
            return

        self._create_fig_ax(fig=fig, ax=ax)

        try:
            voronoi_df = self.convergence_sph_grid.get_spherical_voronoi_areas()
            sns.lineplot(data=voronoi_df, x="N", y="sph. Voronoi cell area", errorbar="sd", ax=self.ax)
            sns.scatterplot(data=voronoi_df, x="N", y="sph. Voronoi cell area", alpha=0.8, color="black", ax=self.ax,
                            s=1)
            sns.scatterplot(data=voronoi_df, x="N", y="ideal area", color="black", marker="x", ax=self.ax)
        except AttributeError:
            pass

        if save:
            self.ax.set_xscale("log")
            self._save_plot_type("voronoi_area_conv")

    def make_spheregrid_time_plot(self, ax=None, fig=None, save=True):
        self._create_fig_ax(fig=fig, ax=ax)
        time_df = self.convergence_sph_grid.get_generation_times()

        sns.lineplot(time_df, x="N", y="Time [s]", ax=self.ax)

        if save:
            self._save_plot_type("spheregrid_time")


class PanelConvergenceSphereGridPlots(PanelRepresentationCollection):

    def __init__(self, dim=3, N_set: list = None, **kwargs):
        list_plots = []
        for alg in GRID_ALGORITHMS[:-1]:
            conv_sphere_grid = ConvergenceSphereGridFactory(alg_name=alg, N_set=N_set, dimensions=dim,
                                                            filter_non_unique=True, **kwargs)
            sphere_plot = ConvergenceSphereGridPlot(conv_sphere_grid)
            list_plots.append(sphere_plot)
        data_name = f"all_convergence_{dim}d"
        super().__init__(data_name, list_plots)

    def make_all_voronoi_area_plots(self, save=True):
        self._make_plot_for_all("make_voronoi_area_conv_plot")
        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots])
        self.set_log_scale(x_axis=True, y_axis=False)
        self.unify_axis_limits()
        if save:
            self._save_multiplot("voronoi_area")

    def make_all_spheregrid_time_plots(self, save=True):
        self._make_plot_for_all("make_spheregrid_time_plot")
        self.add_titles(list_titles=[subplot.get_possible_title() for subplot in self.list_plots])
        self.unify_axis_limits()
        if save:
            self._save_multiplot("spheregrid_times")


if __name__ == "__main__":
    from molgri.constants import SMALL_NS


    # sg = SphereGridFactory().create("ico", 42, 3)
    # sgp = SphereGridPlot(sg, default_complexity_level="half_empty")
    # sgp.make_grid_plot()
    # sgp.make_spherical_voronoi_plot(ax=sgp.ax, fig=sgp.fig, animate_rot=True)

    # conv_obj = ConvergenceSphereGridFactory("ico", dimensions=3, use_saved=True, N_set=DEFAULT_NS)
    # conv_plot = ConvergenceSphereGridPlot(conv_obj)
    # conv_plot.make_voronoi_area_conv_plot(save=True)
    # conv_plot.ax.set_xscale("log")

    # sgp.make_grid_plot()
    # sgp.make_rot_animation(dpi=200)
    # sgp.make_trans_animation()

    psgp = PanelSphereGridPlots(544, 4, default_complexity_level="half_empty", default_context="talk", use_saved=True)
    #psgp.make_all_grid_plots()
    #psgp.make_all_uniformity_plots()
    psgp.make_all_convergence_plots()
    #
    #psgp = PanelSphereGridPlots(700, 3, default_complexity_level="half_empty", default_context="talk", use_saved=True)
    #psgp.make_all_grid_plots(animate_rot=False)
    #psgp.make_all_uniformity_plots()
    #psgp.make_all_convergence_plots()
    #
    # pcsgp = PanelConvergenceSphereGridPlots(3, use_saved=False, N_set=SMALL_NS, filter_non_unique=True)
    # pcsgp.make_all_voronoi_area_plots()

    # for alg in GRID_ALGORITHMS[:-1]:
    # sg = SphereGridFactory.create("cube4D", 200, 3, use_saved=False, filter_non_unique=True)
    # sgp = SphereGridPlot(sg)
    # sgp.make_rot_animation()
    # sgp.make_spherical_voronoi_plot(animate_rot=True)

